import logging
import os
import shutil
from typing import Tuple, List, Union
from calendar import monthrange
import uuid
import requests
import pandas as pd
import numpy as np
from datetime import datetime, date
import time
from tqdm.auto import tqdm
from CommonTools.settings import SETTINGS


def str2dt(dt_str: Union[str, date], format = "%Y-%m-%d") -> date:
    if isinstance(dt_str, str):
        return datetime.strptime(dt_str, format).date()
    if isinstance(dt_str, date):
        return dt_str
    if isinstance(dt_str, datetime):
        return dt_str.date()
    else:
        return dt_str

def dt2str(dt: Union[str, date], format = "%Y-%m-%d") -> str:
    if isinstance(dt, str):
        return dt
    if isinstance(dt, (date, datetime)):
        return dt.strftime(format)
    else:
        return dt
    
def get_month_end_date(snap_year:int, snap_month:int) -> date:
    month_end_date = date(
        snap_year, snap_month,
        monthrange(snap_year, snap_month)[1]
    )
    return month_end_date

def save_xls(names, list_dfs, xls_path):
    with pd.ExcelWriter(xls_path) as writer:
        for n, df in enumerate(list_dfs):
            print(names[n])
            df.to_excel(writer, names[n], index = False)
        #             df.to_excel(writer,'sheet%s' % n, index=False)
        writer.save()


def read_sql_from_file(file: str) -> str:
    with open(file) as obj:
        sql = obj.read()
    return sql

def complie_sql(sql : str, *args) -> str:
    # fill args into sql templates with ?
    arg_num_sql = sql.count("?")
    if len(args) != arg_num_sql:
        raise ValueError(f"SQL temp contains {arg_num_sql} placeholders(?) while only {len(args)} parsed")
    return sql.replace("?", "'{}'").format(*args)

def render_sql_from_file(file: str, **fill_params) -> str:
    with open(file) as obj:
        sql = obj.read()
    if fill_params:
        sql = sql.format(**fill_params)
    return sql

def df_strip(data):
    """
    Remove leading and trailing blank for each column

    Parameters:
    data : DataFrame

    Returns : DataFrame
    """
    for var in data.columns:
        if data[var][data[var].notna()].shape[0] == 0:
            pass
        else:
            if type(data[var][data[var].notna()].iloc[0]) is str:
                data[var] = data[var].str.strip()
    return data



def missing_val_chk(data, var):
    if data[var].isnull().any(axis=None):
        print('There are ' + str(data[var].isnull().sum()) + ' missing values for column ' + var + '.')
        string1 = ','.join(['There are ', str(data[var].isnull().sum()), ' missing values for column ', var, '.'])
        #string1=" "
    else:
        print('Column ' + var + ' has no missing value.')
        string1 = 'Column ' + var + ' has no missing value.'
        #string1=" i"
    return string1


def download_jdbc_jar(jar_file_url, jdbc_driver_folder, jar_file_name: str = None):
    if not os.path.exists(jdbc_driver_folder):
        os.makedirs(jdbc_driver_folder)

    if not jar_file_name:
        jar_file_name = jar_file_url.split('/')[-1]

    jar_path = os.path.join(jdbc_driver_folder, jar_file_name)
    with requests.get(
            jar_file_url,
            # proxies = {
            #     'http': SETTINGS.HTTP_PROXY,
            #     'https': SETTINGS.HTTPS_PROXY
            # }
    ) as response, open(jar_path, 'wb') as out_file:
        out_file.write(response.content)

    logging.info(f"Successfully download JDBC jar from {jar_file_url} and saved to {jar_path}")
    return jar_path


def download_jdbc_jars(jdbc_driver_folder):
    jars = [
        {
            'db_provider' : 'db2',
            'jar_file_url' : 'https://repo1.maven.org/maven2/com/ibm/db2/jcc/db2jcc/db2jcc4/db2jcc-db2jcc4.jar',
            'jar_file_name' : 'db2jcc-db2jcc4.jar'
        },
        {
            'db_provider' : 'mysql',
            'jar_file_url' : 'https://repo1.maven.org/maven2/mysql/mysql-connector-java/8.0.27/mysql-connector-java-8.0.27.jar',
            'jar_file_name' : 'mysql-connector-java.jar'
        },
        {
            'db_provider' : 'sqlite',
            'jar_file_url' : 'https://repo1.maven.org/maven2/org/xerial/sqlite-jdbc/3.36.0.3/sqlite-jdbc-3.36.0.3.jar',
            'jar_file_name' : 'sqlite-jdbc.jar'
        },
        {
            'db_provider' : 'sqlserver',
            'jar_file_url' : 'https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/8.4.0.jre8/mssql-jdbc-8.4.0.jre8.jar',
            'jar_file_name' : 'mssql-jdbc-8.4.0.jre8.jar'
        },
        {
            'db_provider' : 'oracle',
            'jar_file_url' : 'https://repo1.maven.org/maven2/com/oracle/database/jdbc/ojdbc8/19.3.0.0/ojdbc8-19.3.0.0.jar',
            'jar_file_name' : 'ojdbc.jar'
        },
        {
            'db_provider' : 'presto',
            'jar_file_url' : 'https://repo1.maven.org/maven2/io/prestosql/presto-jdbc/347/presto-jdbc-347.jar',
            'jar_file_name' : 'presto.jar'
        },
        {
            'db_provider' : 'postgresql',
            'jar_file_url' : 'https://repo1.maven.org/maven2/org/postgresql/postgresql/42.2.18/postgresql-42.2.18.jar',
            'jar_file_name' : 'postgresql.jar'
        },
        {
            'db_provider': 'sas',
            'jar_file_url': 'https://repos.spark-packages.org/saurfang/spark-sas7bdat/3.0.0-s_2.12/spark-sas7bdat-3.0.0-s_2.12.jar',
            'jar_file_name': 'spark-sas.jar'
        },
        {
            'db_provider': 'parso',  # use to read sas dataset
            'jar_file_url': 'https://repo1.maven.org/maven2/com/epam/parso/2.0.14/parso-2.0.14.jar',
            'jar_file_name': 'parso.jar'
        }
    ]

    if not os.path.exists(jdbc_driver_folder):
        os.makedirs(jdbc_driver_folder)

    for jar in jars:
        db_provider = jar.get('db_provider')
        jar_file_url = jar.get('jar_file_url')
        jar_file_name = jar.get('jar_file_name')

        download_jdbc_jar(jar_file_url, jdbc_driver_folder, jar_file_name)


def readExcel(file, sheetName: str = None, dtype = "str", na_filter: bool = True):
    if file.endswith(".xlsx") or file.endswith(".xls"):
        try:
            with pd.ExcelFile(file) as reader:
                try:
                    file = pd.read_excel(reader, sheetname = sheetName, engine = "python", encoding = "utf8", dtype = dtype, na_filter = na_filter )
                except:
                    file = pd.read_excel(reader, sheet_name = sheetName, dtype = dtype, na_filter = na_filter)
        except Exception as e:
            logging.critical(e)
            file = pd.DataFrame()
    return file


def compare_prod_test(prod_df: pd.DataFrame, test_df: pd.DataFrame):
    """Compare prod dataframe and test dataframe on missing value and matching rate by feature

    :param prod_df: prod version df, benchmark, make sure the joining index is set (single or multiindex)
    :param test_df: test version df, challenger, make sure the joining index is set (single or multiindex)
    :return:
    """
    features = pd.Index(prod_df.columns).intersection(test_df.columns)
    print(f"Size Check:")
    print(f"\tprod_df: {prod_df.shape}  vs. test_df: {test_df.shape}")
    print(f"\t{len(features)} common features found: {list(features)}")
    compare_df = pd.merge(
        prod_df[features],
        test_df[features],
        how='inner',
        left_index=True,
        right_index=True,
        suffixes=('_PROD', '_TEST')
    )
    print(
        f"\t{len(compare_df)} common records found, match rate: prod_df {len(compare_df) / len(prod_df) :.1%}, test_df {len(compare_df) / len(test_df) :.1%}")

    print("\nMatch Rate By Feature")
    # 1.Numerical Feature Check
    numeric_compare = compare_df.select_dtypes(include=['int32', 'int64', 'float32', 'float64'])
    numeric_features = numeric_compare.columns.str.replace('_PROD', '').str.replace('_TEST', '').drop_duplicates(
        keep='first')

    # 2.Categorical Feature Check
    categ_compare = compare_df.select_dtypes(include=['object'])
    categ_features = categ_compare.columns.str.replace('_PROD', '').str.replace('_TEST', '').drop_duplicates(
        keep='first')

    # 3.Datetime Feature Check
    dt_compare = compare_df.select_dtypes(include=['datetime64'])
    dt_features = dt_compare.columns.str.replace('_PROD', '').str.replace('_TEST', '').drop_duplicates(keep='first')

    matching = pd.DataFrame(index=features, columns=['DTYPE', 'NA_PROD', 'NA_TEST', 'MATCH', 'MATCH_VALID'])
    unmatching = {}

    for f in tqdm(numeric_features):
        matching.loc[f, 'DTYPE'] = 'NUM'
        matching.loc[f, 'MATCH'] = np.isclose(compare_df[f + "_PROD"], compare_df[f + "_TEST"]).sum() / len(compare_df)
        matching.loc[f, 'NA_PROD'] = compare_df[f + "_PROD"].isnull().sum() / len(compare_df)
        matching.loc[f, 'NA_TEST'] = compare_df[f + "_TEST"].isnull().sum() / len(compare_df)
        # DROP NA MATCH
        compare_df_valid = compare_df.dropna(subset = [f + "_PROD", f + "_TEST"])
        matching.loc[f, 'MATCH_VALID'] = np.isclose(compare_df_valid[f + "_PROD"], compare_df_valid[f + "_TEST"]).sum() / len(compare_df_valid)

        unmatching[f] = compare_df.loc[
            ~np.isclose(compare_df[f + "_PROD"], compare_df[f + "_TEST"]), [f + "_PROD", f + "_TEST"]]

    for f in tqdm(categ_features):
        matching.loc[f, 'DTYPE'] = 'CATEG'
        matching.loc[f, 'MATCH'] = (compare_df[f + "_PROD"] == compare_df[f + "_TEST"]).sum() / len(compare_df)
        matching.loc[f, 'NA_PROD'] = compare_df[f + "_PROD"].isnull().sum() / len(compare_df)
        matching.loc[f, 'NA_TEST'] = compare_df[f + "_TEST"].isnull().sum() / len(compare_df)
        # DROP NA MATCH
        compare_df_valid = compare_df.dropna(subset=[f + "_PROD", f + "_TEST"])
        matching.loc[f, 'MATCH_VALID'] = (compare_df_valid[f + "_PROD"] == compare_df_valid[f + "_TEST"]).sum() / len(compare_df_valid)

        unmatching[f] = compare_df.loc[compare_df[f + "_PROD"] != compare_df[f + "_TEST"], [f + "_PROD", f + "_TEST"]]

    for f in tqdm(dt_features):
        matching.loc[f, 'DTYPE'] = 'CATEG'
        matching.loc[f, 'MATCH'] = (pd.to_datetime(compare_df[f + "_PROD"]) == pd.to_datetime(
            compare_df[f + "_TEST"])).sum() / len(compare_df)
        matching.loc[f, 'NA_PROD'] = compare_df[f + "_PROD"].isnull().sum() / len(compare_df)
        matching.loc[f, 'NA_TEST'] = compare_df[f + "_TEST"].isnull().sum() / len(compare_df)
        # DROP NA MATCH
        compare_df_valid = compare_df.dropna(subset=[f + "_PROD", f + "_TEST"])
        matching.loc[f, 'MATCH_VALID'] = (pd.to_datetime(compare_df_valid[f + "_PROD"]) == pd.to_datetime(
            compare_df_valid[f + "_TEST"])).sum() / len(compare_df_valid)

        unmatching[f] = compare_df.loc[
            pd.to_datetime(compare_df[f + "_PROD"]) != pd.to_datetime(compare_df[f + "_PROD"]), [f + "_PROD",
                                                                                                 f + "_TEST"]]

    matching.sort_values(by=['MATCH'], ascending=[False], inplace=True)
    return matching, unmatching

def error_analysis(unmatching_f: pd.DataFrame, dtype: str):
    """

    :param unmatching_f:
    :param dtype:{INT, FLOAT, DATE, CATEG}
    :return:
    """

    un = unmatching_f.copy()
    if dtype == 'INT':
        un['DIFF'] = un.iloc[:, 0] - un.iloc[:, 1]
        return un['DIFF'].value_counts(normalize = True, dropna = False)
    if dtype == 'FLOAT':
        un['DIFF'] = un.iloc[:, 0] - un.iloc[:, 1]
        return un.describe()
    if dtype == 'CATEG':
        from sklearn.metrics import confusion_matrix

        value_categs = un.iloc[:, 0].value_counts().index
        return pd.DataFrame(
            confusion_matrix(un.iloc[:, 0], un.iloc[:, 1], labels = value_categs),
            columns = value_categs,
            index = value_categs
        ).style.applymap(lambda x: "background-color: #F28C77" if x > 0 else "background-color: white")

def loop_timer(total_steps: int) -> Tuple[int, float, float, float, float]:
    _start = time.time()
    for step in range(total_steps):
        _lap = time.time()
        progress = (step + 1) / total_steps
        time_step = _lap - _start
        time_elasped = _lap - _start
        time_remaining = time_elasped / (step + 1) * (total_steps - step - 1)
        yield step + 1, progress, time_step, time_elasped, time_remaining


def convert_np_dtype(df: pd.DataFrame, exclude_cols: List[str] = None, keep_nonnull_int: bool = True, keep_date:bool = True) -> pd.DataFrame:
    """convert dtype from pandas to numpy compatible (sklearn), i.e., only object, datetime, int and float type

    :param df:
    :param exclude_cols: columns to exclude from conversion
    :param keep_nonnull_int: if True, will convert non-null nullable Integer to numpy non-nullable integer,
            otherwise will convert nullable Integer to float
    :param keep_date: if False, will convert datetime type to date string
    :return:
    """

    if exclude_cols is None:
        exclude_cols = []
    # convert nullable integer
    dtypes = df.dtypes
    nullable_int_cols = dtypes[
        (dtypes.apply(lambda d: d.name).str.startswith("Int"))
        | (dtypes.apply(lambda d: d.name).str.startswith("UInt"))
        ].index

    non_null_int_cols = []
    if keep_nonnull_int:
        # will convert non-null nullable Integer to numpy non-nullable integer
        null_int_stat = df[nullable_int_cols].isna().sum()
        non_null_int_cols = null_int_stat[null_int_stat == 0].index
        nullable_int_cols = nullable_int_cols.difference(non_null_int_cols)

    if not keep_date:
        for col in df.select_dtypes(include=['datetime']):
            df[col] = df[col].dt.date.apply(dt2str)

    return df.astype(
        {
            c: 'object'
            for c in df.select_dtypes(exclude=['number', 'datetime']).columns
            if c not in exclude_cols
        }
    ).astype(
        {
            c: 'float'
            for c in nullable_int_cols
            if c not in exclude_cols
        }
    ).astype(
        {
            c: 'int'
            for c in non_null_int_cols
            if c not in exclude_cols
        }
    )


class TempFolder:
    def __init__(self, root_path: str):
        folder = str(uuid.uuid4())
        self.temp_folder = os.path.join(root_path, folder)

    def __enter__(self):
        if not os.path.exists(self.temp_folder):
            os.makedirs(self.temp_folder)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.temp_folder, ignore_errors = True)

    def getTempFolderPath(self):
        return self.temp_folder
    
if __name__ == '__main__':
    dt = date(2021, 1, 4)
    print(dt2str(dt))





import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone, is_classifier, is_regressor
from sklearn.metrics import get_scorer
from sklearn.preprocessing import LabelEncoder


# https://www.statsmodels.org/stable/regression.html#
# https://www.statsmodels.org/stable/generated/statsmodels.regression.linear_model.RegressionResults.html#statsmodels.regression.linear_model.RegressionResults
# https://stats.stackexchange.com/questions/16493/difference-between-confidence-intervals-and-prediction-intervals
# https://stats.stackexchange.com/questions/136157/general-mathematics-for-confidence-interval-in-multiple-linear-regression

class _linearStatsModelWrapper(BaseEstimator):
    def __init__(self, model_family: str = 'glm', fit_intercept: bool = False, model_params: dict = None,
                 fit_params: dict = None, scoring: str = None):
        self.model_family = model_family
        self.fit_intercept = fit_intercept
        self.model_params = {} if model_params is None else model_params
        self.fit_params = {} if fit_params is None else fit_params
        self.scoring = scoring
        self.le = None

    def fit(self, X, y, sample_weight=None):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        self.feature_names_in_ = X.columns
        model_cls = {
            # continious regression model
            'ols': sm.OLS,
            'gls': sm.GLS,
            'wls': sm.WLS,
            'rlm': sm.RLM,
            # discrete classification model
            'glm': sm.GLM,
            'logit': sm.Logit,
            'probit': sm.Probit,
            'poisson': sm.Poisson,
        }.get(self.model_family, sm.GLM)

        if self.fit_intercept:
            X = sm.add_constant(X)
        if self.le is not None:
            self.le = self.le.fit(y)
            y = self.le.transform(y)

        model = model_cls(y, X, **self.model_params)
        self.model_result = model.fit(freq_weight=sample_weight, **self.fit_params)

        return self

    def score(self, X, y, sample_weight=None):
        if callable(self.scoring):
            s = self.scoring(self, X, y)
            return s

        try:
            s = getattr(self.model_result_, self.scoring)
        except AttributeError:
            if isinstance(self.scoring, str):
                scorer = get_scorer(self.scoring)
                s = scorer(self, X, y)
            else:
                raise TypeError("Scoring must be str or callable")
        return s

    @property
    def model_result_(self):
        return self.model_result


class LinearClfStatsModelWrapper(_linearStatsModelWrapper, ClassifierMixin):
    def __init__(self, model_family: str = 'glm', fit_intercept: bool = False, model_params: dict = None,
                 fit_params: dict = None, scoring: str = 'roc_auc'):
        if model_family in ('ols', 'gls', 'wls', 'rlm'):
            raise ValueError(
                "the model_family falls in the regresion type model, please use LinearRegStatsModelWrapper class instead")
        super().__init__(model_family=model_family, fit_intercept=fit_intercept, model_params=model_params,
                         fit_params=fit_params, scoring=scoring)
        self.le = LabelEncoder()

    def decision_function(self, X) -> np.ndarray:
        if len(self.classes_) > 2:
            return self.predict_proba(X)
        else:
            return self.predict_proba(X)[:, 1]

    def predict_proba(self, X) -> np.ndarray:
        if self.fit_intercept:
            X = sm.add_constant(X)
        probs = self.model_result_.predict(X).values
        return np.c_[1 - probs, probs]

    def predict_log_proba(self, X) -> np.ndarray:
        return np.log(self.predict_proba(X))

    def predict(self, X) -> np.ndarray:
        probs = self.predict_proba(X)[:, 1]
        return np.where(probs >= 0.5, 1, 0)

    @property
    def classes_(self):
        return self.le.classes_

    @property
    def coef_(self):
        coefs = self.model_result_.params
        if self.fit_intercept:
            return coefs.values[1:]  # 1st is intercept
        else:
            return coefs.values

    @property
    def intercept_(self):
        coefs = self.model_result_.params
        if self.fit_intercept:
            return np.array([coefs['const']])
        else:
            return np.array([0.0])


class LinearRegStatsModelWrapper(_linearStatsModelWrapper, RegressorMixin):
    def __init__(self, model_family: str = 'ols', fit_intercept: bool = False, model_params: dict = None,
                 fit_params: dict = None, scoring: str = 'r2'):
        if model_family not in ('ols', 'gls', 'wls', 'rlm'):
            raise ValueError(
                "the model_family falls in the classification type model, please use LinearClfStatsModelWrapper class instead")
        super().__init__(model_family=model_family, fit_intercept=fit_intercept, model_params=model_params,
                         fit_params=fit_params, scoring=scoring)

    def predict(self, X) -> np.ndarray:
        if self.fit_intercept:
            X = sm.add_constant(X)
        return self.model_result_.predict(X).values



from __future__ import annotations
from datetime import date
import logging
from typing import List
import ibis
from ibis import Schema
import pandas as pd
import pyspark
from clickhouse_connect.driver.tools import insert_file
from CommonTools.sparker import SparkConnector
from CommonTools.utils import dt2str, str2dt
from CommonTools.dtyper import DSchema
from CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from ProviderTools.clickhouse.dbapi import ClickHouse

class SnapshotDataManagerCHSQL(SnapshotDataManagerBase):
    """files are saved as clickhouse tables under each schema.table
    """
    @classmethod
    def setup(cls, db_conf: ClickHouse, spark_connector: SparkConnector = None, **settings):
        super(SnapshotDataManagerCHSQL, cls).setup(spark_connector = spark_connector)
        cls._db_conf = db_conf
        cls._ops = ibis.clickhouse.connect(
            user=db_conf.username,
            password=db_conf.password,
            host=db_conf.ip,
            port=db_conf.port,
            **settings
        )
        
    def __init__(self, schema:str, table:str, snap_dt_key: str):
        """database management interface

        :param schema: schema
        :param table:  table under each schema
        :param snap_dt_key: snap date column name for all tables
        """
        super().__init__(schema = schema, table = table)
        self.snap_dt_key = snap_dt_key
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        if self.schema not in self._ops.list_databases(like = self.schema):
            return False
        return self.table in self._ops.list_tables(like = self.table, database = self.schema)

    def init_table(self, col_schemas: DSchema, overwrite:bool = False, **settings):
        """initialize/create table in the underlying data warehouse system

        :param DSchema col_schemas: data column schema
        :param List[str] primary_keys: primary keys, defaults to None
        :param bool overwrite: whether to drop table if exists, defaults to False
        """

        if self.exist():
            if overwrite:
                self.drop()
            else:
                logging.warning(f"{self.schema}.{self.table} already exists, will do nothing." 
                                "set overwrite to True if you wish to reset table")
                return
        
        # create schema
        self._ops.create_database(
            name = self.schema,
            force = True
        )
        # create table
        primary_keys = col_schemas.primary_keys
        partition_keys = col_schemas.partition_keys
        if self.snap_dt_key not in partition_keys:
            partition_keys.append(self.snap_dt_key)
        logging.info(f"Creating table {self.schema}.{self.table} using schema:\n{col_schemas.ibis_schema}")
        logging.info(f"primary keys = {primary_keys}, partition_keys = {partition_keys}")
        self._ops.create_table(
            name = self.table,
            schema = col_schemas.ibis_schema,
            database = self.schema,
            engine = "MergeTree",
            order_by = primary_keys,
            partition_by = partition_keys,
            settings = settings
        )
        # add column descriptions
        descrs = col_schemas.descrs
        for col, descr in descrs.items():
            try:
                sql = f"""
                ALTER TABLE {self.schema}.{self.table} 
                COMMENT COLUMN IF EXISTS {col} '{descr}'"""
                self._ops.con.command(sql)
            except:
                pass
        
    def get_schema(self) -> Schema:
        """Check dtypes of the given schema/dataset

        :return:
        """
        return self._ops.get_schema(
            table_name = self.table,
            database = self.schema
        )
        
    def count(self, snap_dt: date) -> int:
        if not self.exist():
            return 0
        table = self._ops.table(
            name = self.table,
            database = self.schema
        )
        return table.count(
            where = (table[self.snap_dt_key] == snap_dt)
        ).to_pandas() # convert to a scalar number
        
    def save_qry(self, query: str, snap_dt: date, overwrite: bool = False):
        """save the query into the table using "insert into schema.table select ... " clause

        :param query: the sql query for table to be inserted
        :param snap_dt: the snap date partition to be inserted, use to check existence
        :param overwrite: whether to overwrite the table if exists

        :return:
        """
        # first detect whether snap date already exists
        clear = self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite)
        if clear is False:
            return

        # first create a view of that table which can be inspected the column dtypes and names
        qry_cols = self._ops.sql(query).columns

        # insert into the table
        # clickhouse insert need column order correct
        sql = f"""
        INSERT INTO {self.schema}.{self.table} ({','.join(qry_cols)})
        {query}
        """
        logging.info(f"Inserting into {self.schema}.{self.table}@{snap_dt} using query:\n{sql}")
        self._ops.con.command(sql)

        logging.info(f"Successfully saved to {self.schema}.{self.table}@{snap_dt}")
        
    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        self._ops.insert(
            name = self.table,
            obj = df,
            database = self.schema,
            **kws
        )
        
    def ingest_from_file(self, file_path:str, snap_dt: date, header:bool = True, overwrite:bool = True):
        if overwrite:
            self.delete(snap_dt = snap_dt)
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"{num_records} records found for {self.schema}.{self.table}@{snap_dt}, will do nothing")
                return
            
        # extract file format
        if file_path.endswith("parquet"):
            fmt = "Parquet"
        elif file_path.endswith("csv"):
            fmt = "CSVWithNames" if header else "CSV"
        else:
            raise TypeError("File not supported")
        
        insert_file(
            client = self._ops.con,
            table = self.table,
            database = self.schema,
            file_path = file_path,
            fmt = fmt,
        )
        logging.info(f"Successfully ingested file {file_path} to {self.schema}.{self.table}")
        
    def get_existing_snap_dts(self) -> List[date]:
        if not self.exist():
            return []
        existing_snaps = (
            self._ops.table(
                name = self.table,
                database = self.schema
            ).select(self.snap_dt_key)
            .distinct()
            .order_by(self.snap_dt_key)
            .to_pandas()
        )
        if len(existing_snaps) == 0:
            return []
        return list(
            str2dt(dt.date()) 
            for dt in pd.to_datetime(existing_snaps[self.snap_dt_key]).dt.to_pydatetime()
        )
        
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        table = self._ops.table(
            name = self.table,
            database = self.schema
        )
        df = table.filter(table[self.snap_dt_key] == snap_dt)
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df.to_pandas()
    
    def reads(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        table = self._ops.table(
            name = self.table,
            database = self.schema
        )
        df = table.filter(table[self.snap_dt_key].isin(snap_dts))
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df.to_pandas()
    
    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} = '{snap_dt}'
        """
        if hasattr(self, "sc"):
            df = self.sc.query_db(self._db_conf, sql)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def loads(self, snap_dts: List[date], **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        if 'columns' in kws:
            cols = ','.join(kws['columns'])
        else:
            cols = '*'
        snap_dt_range = ",".join(f"'{dt}'" for dt in snap_dts)
        sql = f"""
        select {cols} from {self.schema}.{self.table} where {self.snap_dt_key} in [{snap_dt_range}]
        """
        if hasattr(self, "sc"):
            df = self.sc.query_db(self._db_conf, sql)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")
        
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        if self.exist():
            sql = f"""
            ALTER TABLE {self.schema}.{self.table} DELETE
            WHERE {self.snap_dt_key} = '{snap_dt}'
            """
            logging.info(f"Deleting table {self.schema}.{self.table} using query:\n{sql}")
            self._ops.con.command(cmd = sql)
        
    def drop(self):
        """drop the whole table

        :return:
        """
        self._ops.drop_table(
            name = self.table,
            database = self.schema,
            force = True
        )
        
    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerCHSQL:
        sql = """
        insert into {dst_schema:Identifier}.{dst_table:Identifier} 
        select * from {src_schema:Identifier}.{src_table:Identifier}"""
        args = dict(
            dst_schema = dst_schema, 
            dst_table = dst_table, 
            src_schema = self.schema, 
            src_table = self.table
        )
        self._ops.con.command(cmd = sql, parameters = args)
        new = SnapshotDataManagerCHSQL(
            schema = dst_schema,
            table = dst_table,
            snap_dt_key = self.snap_dt_key
        )
        return new
    
    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        sql = """
        select 
            sum(rows) as rows,
            sum(bytes_on_disk) as bytes_on_disk
        from system.parts
        where 
            active
            and database = %(schema)s
            and table = %(table)s
            and partition = %(snap_dt)s
        """
        if not self.exist():
            return 0
        
        args = dict(schema = self.schema, table =self.table, snap_dt = snap_dt)
        d = self._ops.con.query(query=sql, parameters=args).first_item

        size_bytes, rows = d['bytes_on_disk'], d['rows']
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size


'''
parallel run example:
def slow_power(x, p):
    time.sleep(1)
    return x ** p
from joblib import Parallel, delayed
number_of_cpu = joblib.cpu_count()
delayed_funcs = [delayed(slow_power)(i, 5) for i in range(10)]
parallel_pool = Parallel(n_jobs=number_of_cpu)
parallel_pool(delayed_funcs)
+> equivalent to:
def delayer(gunc):
    gunc = delayed(gunc)
    def wrapper(*args, **kws):
        return gunc(*args, **kws)
    return wrapper
def parallel_runner(jobs: iterator, n_jobs = -1):
    return Parallel(n_jobs=n_jobs)(jobs)
@ delayer
def slow_power2(x, p):
    time.sleep(1)
    return x + p, x - p
jobs = (slow_power2(i, 5) for i in range(20))
parallel_runner(jobs)
'''
from joblib import delayed, Parallel


def delayer(gunc):
    # the decorator for delay running funcs, see examples above
    gunc = delayed(gunc)

    def wrapper(*args, **kws):
        return gunc(*args, **kws)

    return wrapper


def parallel_run(jobs, n_jobs: int = -1) -> list:
    """
    :param jobs: iterator of jobs to run
    :param n_jobs: number of cpus to run, -1 = use all available core
    :return: [(results from job1), (results from job2), (results from job3)]
    """
    return Parallel(n_jobs=n_jobs)(jobs)


import logging
import os
import subprocess
from typing import List, Tuple
import clickhouse_connect
from clickhouse_connect.driver.tools import insert_file
import pandas as pd

from CommonTools.dbapi import baseDbInf
from CommonTools.dtyper import Schema

class ClickHouse(baseDbInf):
    def __init__(self, driver="clickhouse+native"):
        super().__init__(driver)

    def argTempStr(self):
        return "%s"

    def getJdbcUrl(self) -> str:
        return f"jdbc:clickhouse://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/github/housepower/clickhouse-native-jdbc-shaded/2.6.5/clickhouse-native-jdbc-shaded-2.6.5.jar
        return "com.github.housepower.jdbc.ClickHouseDriver"

    def getConnStr(self) -> str:
        # only used for sqlalchemy driver type
        return f"{self.driver}://{self.username}:{self.password}@{self.ip}:{self.port}"


class ClickhouseSchema:
    def __init__(self, dtype_dict: dict, default_dict: dict = None, note_dict: dict = None):
        self.dtype_dict = dtype_dict
        self.default_dict = default_dict if default_dict is not None else {}
        self.note_dict = note_dict if note_dict is not None else {}

    def to_dataframe(self) -> pd.DataFrame:
        s = pd.merge(
            pd.Series(self.dtype_dict, dtype='string').rename("dtype"),
            pd.Series(self.default_dict, dtype='string').rename("default"),
            how='left',
            right_index=True,
            left_index=True
        )
        s = pd.merge(
            s,
            pd.Series(self.note_dict, dtype='string').rename("note"),
            how='left',
            right_index=True,
            left_index=True
        )
        return s

    @classmethod
    def from_dataframe(cls, schema_df: pd.DataFrame):
        return cls(
            dtype_dict=schema_df['dtype'].to_dict(),
            default_dict=schema_df['default'].dropna().to_dict(),
            note_dict=schema_df['note'].dropna().to_dict(),
        )

    def to_sql(self) -> str:
        l = []
        for col, dtype in self.dtype_dict.items():
            base = f"{col} {dtype}"
            if self.default_dict.get(col) is not None:
                default_v = self.default_dict.get(col)
                base += f" DEFAULT {default_v}"
            if self.note_dict.get(col) is not None:
                base += f" COMMENT '{self.note_dict.get(col)}'"

            l.append(base)

        return ",\n".join(l)


class ClickhouseCRUD:
    @classmethod
    def setup(cls, base_path: str):
        cls.base_path = base_path

    def __init__(self, ch_conf: ClickHouse):
        self.client = clickhouse_connect.get_client(
            host=ch_conf.ip,
            port=ch_conf.port,
            username=ch_conf.username,
            password=ch_conf.password
        )

    def is_exist(self, schema: str, table: str) -> bool:
        sql = "EXISTS {schema:Identifier}.{table:Identifier}"
        args = dict(schema = schema, table = table)
        r = self.client.command(cmd = sql, parameters = args) # will return single value for simple query
        return r

    def extract_file_suffix(self, file_path: str, header: bool = True) -> str:
        if file_path.endswith("parquet"):
            fmt = "Parquet"
        elif file_path.endswith("csv"):
            fmt = "CSVWithNames" if header else "CSV"
        else:
            raise TypeError("File not supported")
        return fmt
    
    def show_tables(self, schema: str) -> List[str]:
        sql = "SELECT DISTINCT name FROM system.tables WHERE database={schema:String} AND has_own_data = 1"
        args = dict(schema = schema)
        df = self.client.query_df(query = sql, parameters = args)
        return df['name'].tolist()

    def create_schema(self, schema: str):
        sql = "CREATE DATABASE IF NOT EXISTS {schema:Identifier}"
        args = dict(schema = schema)
        self.client.command(cmd = sql, parameters = args)
        logging.info(f"Successfully created schema {schema}")

    def detect_file_schema(self, file_path: str, header: bool = True) -> pd.DataFrame:
        fmt = self.extract_file_suffix(file_path=file_path, header=header)
        sql = "DESCRIBE TABLE file({file_path: String}, {fmt:Identifier})"
        args = dict(file_path = file_path, fmt = fmt)
        return self.client.query_df(query = sql, parameters = args)

    def detect_table_schema(self, schema: str, table: str) -> pd.DataFrame:
        sql = "DESCRIBE TABLE {schema:Identifier}.{table:Identifier}"
        args = dict(schema=schema, table=table)
        return self.client.query_df(query = sql, parameters = args)

    def mount_file_src(self, src_folder_path: str, dest_folder_name: str = None) -> str:
        if dest_folder_name is None:
            dest_folder_name = os.path.relpath(src_folder_path, os.path.dirname(src_folder_path))
        dest_path = os.path.join(self.base_path, "usr/bin/user_files", dest_folder_name)
        command = f"ln -s {src_folder_path} {dest_path}"
        proc = subprocess.Popen(command, shell=True, close_fds=True)
        return command

    def unmount_file_src(self, dest_folder_name: str) -> str:
        dest_path = os.path.join(self.base_path, "usr/bin/user_files", dest_folder_name)
        command = f"unlink {dest_path}"
        proc = subprocess.Popen(command, shell=True, close_fds=True)
        return command

    def create_table(self, schema: str, table: str, col_schemas: pd.DataFrame,
                     engine: str = 'MergeTree', partition_keys: List[str] = None, primary_keys: List[str] = None,
                     order_keys: List[str] = None,
                     table_note: str = None) -> Tuple[int, str]:

        f = ClickhouseSchema.from_dataframe(col_schemas)

        sql = f"CREATE TABLE IF NOT EXISTS {schema}.{table}" + "\n(\n" + f.to_sql() + "\n)\n" + f"ENGINE = {engine}"
        if partition_keys is not None:
            partition_keys = """PARTITION BY ({})""".format(",".join(partition_keys))
            sql += ("\n" + partition_keys)
        if primary_keys is not None:
            primary_keys = """PRIMARY KEY ({})""".format(",".join(primary_keys))
            sql += ("\n" + primary_keys)
        if order_keys is not None:
            order_keys = """ORDER BY ({})""".format(",".join(order_keys))
            sql += ("\n" + order_keys)
        if table_note is not None:
            table_note = """COMMENT '{}' """.format(table_note)
            sql += ("\n" + table_note)
        return self.client.command(cmd = sql), sql

    def create_view(self, schema: str, view: str, query: str, materialized: bool = False) -> Tuple[int, str]:
        if materialized:
            sql = f"""CREATE OR REPLACE MATERIALIZED VIEW {schema}.{view} AS"""
        else:
            sql = f"""CREATE OR REPLACE VIEW {schema}.{view} AS"""
        sql += f"\n{query}"
        return self.client.command(cmd = sql), sql

    def drop_table(self, schema: str, table: str):
        sql = "DROP TABLE {schema:Identifier}.{table:Identifier}"
        args = dict(schema=schema, table=table)
        self.client.command(cmd=sql, parameters=args)
        logging.info(f"Successfully dropped table {schema}.{table}")

    def drop_view(self, schema: str, view: str):
        sql = 'DROP VIEW {schema:Identifier}.{view:Identifier}'
        args = dict(schema=schema, view=view)
        self.client.command(cmd=sql, parameters=args)
        logging.info(f"Successfully dropped view {schema}.{view}")

    def ingest_file(self, schema: str, table: str, file_path: str, header: bool = True, use_cli: bool = False) -> int:
        fmt = self.extract_file_suffix(file_path=file_path, header=header)
        if use_cli:
            query = f"INSERT INTO {schema}.{table} FROM INFILE '{file_path}' FORMAT {fmt}"
            command = f"""{self.base_path}/usr/bin/clickhouse client --query="{query}" """
            proc = subprocess.Popen(command, shell=True)
            # proc.wait()
            logging.info(f"Submitted file ingestion task for {file_path} to {schema}.{table}")
            return command
        else:
            insert_file(
                self.client,
                table=table,
                database=schema,
                file_path=file_path,
                fmt=fmt,
            )
            logging.info(f"Successfully ingested file {file_path} to {schema}.{table}")
        return 1

    def insert_from_select(self, schema: str, table: str, query: str, qry_cols: list):
        """save the query into the table using "insert into schema.table select ... " clause

        :param schema: schema in which the table exists
        :param table: table to be inserted
        :param query: the sql query for table to be inserted
        :param qry_cols: column order matters because clickhouse use column order for mapping,
                so qry_cols should match query result column order
        :return:
        """
        sql = f"""
        INSERT INTO {schema}.{table} ({','.join(qry_cols)})
        {query}
        """
        self.client.command(cmd = sql)

    def truncate_table(self, schema: str, table: str) -> str:
        sql = f"""TRUNCATE TABLE {schema}.{table}"""
        self.client.command(cmd = sql)
        logging.info(f"Successfully truncated table {schema}.{table}")
        return sql

    def delete_data(self, schema: str, table: str, where_clause: str = None) -> str:
        sql = f"""ALTER TABLE {schema}.{table} DELETE"""
        if where_clause is not None:
            sql += f" WHERE {where_clause}"
        else:
            sql += " WHERE 1=1"
        self.client.command(cmd = sql)
        logging.info(f"Successfully deleted data")
        return sql

    def size_by_partition(self, schema: str, table: str, partition_keys: List[str]) -> pd.DataFrame:
        partitions = ",".join(partition_keys)
        sql = f"""
        SELECT
            {partitions},
            COUNT()
        FROM
            {schema}.{table}
        GROUP BY
            {partitions}
        """
        return self.client.query_df(query = sql)


import os
import io
import glob
from typing import Union, Optional
import logging
import pandas as pd
from google.cloud import storage
from google.oauth2 import service_account

from CommonTools.obj_storage import ObjStorage, remove_begin_slash, add_tail_slash

class GCSAccessor(ObjStorage):
    def __init__(self, sa_config: Optional[dict] = None, project: str = None, region_name: str = None):
        if not sa_config:
            self.client = storage.Client()
        else:
            if os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
                self.client = storage.Client()
            else:
                creds = service_account.Credentials.from_service_account_info(sa_config)
                self.client = storage.Client(credentials=creds)
                
        self.project = project
        self.region_name = region_name
        
    def get_uri_protocol(self) -> str:
        return "gs"
        
    def list_buckets(self) -> pd.DataFrame:
        buckets = self.client.list_buckets()
        props = []
        for bucket in buckets:
            props.append({
                'Name' : bucket._properties["name"],
                'CreationDate' : pd.to_datetime(bucket._properties["timeCreated"])
            })
        return pd.DataFrame.from_records(props)
        
    def create_bucket(self, bucket_name: str) -> int:
        """Create a new bucket

        :param bucket_name: name of the bucket
        :return:
        """
        bucket = self.client.create_bucket(
            bucket_or_name = bucket_name,
            project = self.project,
            location = self.region_name
        )
        return 1
    
    def get_bucket(self, bucket_name: str = None):
        return self.client.get_bucket(bucket_name)
    
    def enter_bucket(self, bucket_name: str = None) -> int:
        """move to a given bucket

        :param bucket_name: name of the bucket you want to switch to
        :return:
        """
        # check if exists
        if self.get_bucket(bucket_name):
            self.BUCKET = bucket_name
            return 1
        else:
            logging.error(f"Bucket {bucket_name} does not exists, please call .list_buckets() to see all available buckets")
            return 0
    
    def if_not_exist_return_default_bucket(self, bucket_name = None):
        return self.BUCKET if bucket_name is None else bucket_name
    
    def get_current_bucket(self):
        """get your current bucket object

        :return:
        """
        """get your current bucket object

        :return:
        """
        if hasattr(self, 'BUCKET'):
            return self.get_bucket(self.BUCKET)
        else:
            raise ValueError("Has not enter into any bucket yet, please call .enter_bucket(bucket_name) to enter a specific bucket")
        
    def delete_bucket(self, bucket_name: str, interactive: bool = True) -> int:
        """delete the bucket

        :param bucket_name: name of the bucket to delete
        :param interactive: whether you would like an interactive warning message
        :return:
        """
        logging.warning(f"You are about to delete bucket: {bucket_name} and everything under it, please confirm")
        if interactive:
            y_n = input(f"delete everything? (y/n)")
            if y_n in ['y', 'Y']:
                self.get_bucket(bucket_name).delete(force=True)
                return 1
            else:
                logging.warning(f"Aborted by user, please try again")
                return 0

        self.resource.Bucket(bucket_name).delete()
        return 1
    
    ''' below requires lock a bucket '''
    def list_objs(self, remote_path: str, bucket_name: str = None) -> pd.DataFrame:
        """list objects under the specific directory

        :param remote_path: remote path excluding bucket
        :param bucket_name: name of the bucket, if not given, will use the default bucket
        :return: pandas dataframe showing files inside
        """
        pd.set_option('display.max_colwidth', None)

        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        remote_path = remove_begin_slash(remote_path)

        try:
            r = []
            for blob in self.client.list_blobs(bucket_or_name=bucket_name, prefix = remote_path):
                r.append({
                    'Key': blob.name,
                    'LastModified' : blob.updated,
                    'Size': blob.size
                })
        except Exception as e:
            logging.error(f'Error occurred when listing objects/files under path: {bucket_name}/{remote_path}: \n{e}')
        else:
            if len(r) == 0:
                print(f"No objects found at {bucket_name}/{remote_path}")
            else:
                df = pd.DataFrame.from_records(r)

                logging.info(f"There are {len(r)} files found in gcs://{bucket_name}/{remote_path}")
                return df[['Key', 'LastModified', 'Size']]
            
    def get_obj(self, remote_path:str, bucket_name: str = None) -> storage.blob.Blob:
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        obj = bucket.get_blob(blob_name = remove_begin_slash(remote_path))
        return obj
    
    def create_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Create an object

        :param remote_path: if remote path ends with /, it will be folder, otherwise will be a blank file
        :param bucket_name:
        :return: 1 if success otherwise the info dictionary
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(remove_begin_slash(remote_path))
        blob.upload_from_string('', content_type=None)
        return 1
        
    
    def delete_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Delete an object/file

        :param remote_path: the filepath to delete (excluding bucket path)
        :param bucket_name: in which bucket
        :return:
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        obj = bucket.get_blob(blob_name = remove_begin_slash(remote_path))
        obj.delete()
        return 1
    
    def create_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Create a folder object on GCS

        :param remote_folder_path: the folder path (excluding bucket path)
        :param bucket_name: under which bucket
        :return:
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        remote_folder_path = add_tail_slash(remove_begin_slash(remote_folder_path))
        response = self.create_obj(remote_folder_path, bucket_name)
        return response
    
    def delete_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Delete the folder and every file underneath it

        :param remote_folder_path: the folder path to delete (excluding bucket path)
        :param bucket_name:
        :return:
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        remote_folder_path = add_tail_slash(remove_begin_slash(remote_folder_path))
        for blob in bucket.list_blobs(prefix=remote_folder_path):
            blob.delete()
        return 1
    
    def copy_obj(self, from_bucket_name: str = None, from_path: str = None, to_bucket_name: str = None, to_path: str = None):
        """Copy one file from one place to another (can copy to a different bucket)

        :param from_bucket_name: from which bucket you want to copy the file
        :param from_path: which file you want to copy (excluding bucket path)
        :param to_bucket_name: to which bucket you want to copy to
        :param to_path: the destination file path (excluding bucket path)
        :return:
        """
        from_bucket_name = self.if_not_exist_return_default_bucket(from_bucket_name)
        to_bucket_name = self.if_not_exist_return_default_bucket(to_bucket_name)
        
        source_bucket = self.client.get_bucket(from_bucket_name)
        dest_bucket = self.client.get_bucket(to_bucket_name)
        
        source_blob = source_bucket.blob(remove_begin_slash(from_path))
        source_bucket.copy_blob(
            blob = source_blob, 
            destination_bucket = dest_bucket, 
            new_name = remove_begin_slash(to_path), 
            if_generation_match = 0,
        )
        
    def upload_file(self, local_path: str, remote_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        '''upload file from local machine to GCS

        :param local_path: local file path that you would like to upload to GCS, suggest to be absolute path
        :param remote_path: remote path on MinIO (excluding bucket path); could either start with / or not
        :param multipart_threshold: Setting an extremely large multipart threshold to effectively disable multipart uplaod.
                Note: Multiplart upload is not compatible with GCS, therefore disabling multipart upload for compatability with hybrid solution.
                set it to None is you want multipart saving
        :return:
        '''
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        try:
            bucket = self.client.get_bucket(bucket_name)
            blob = bucket.blob(key)
            blob.upload_from_filename(local_path)
        except Exception as e:
            logging.error(f'File upload FAILED from {local_path} to {key}:\n'+ str(e))
        else:
            logging.info(f'File uploaded successfully from {local_path} to {key}:')
            
    def download_file(self, remote_path: str, local_path: str, bucket_name: str = None):
        '''download file from GCS to local machine

        :param remote_path: remote path on GCS (excluding bucket); could either start with / or not
        :param local_path: local file path that you would like to save the file from MinIO, suggest to be absolute path
        :param bucket_name
        :return:
        '''
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        try:
            bucket = self.client.get_bucket(bucket_name)
            blob = bucket.blob(key)
            blob.download_to_filename(local_path)
        except Exception as e:
            logging.error(f'File download FAILED from {key} to {local_path}:\n'+ str(e))
        else:
            logging.info(f'File download successfully from {key} to {local_path}:')
            
    def read_obj(self, remote_path: str, bucket_name: str = None) -> bytes:
        """read obj from GCS without downloading

        :param remote_path:  remote path to read on GCS (excluding bucket); could either start with / or not
        :param Bucket:
        :return: a buffer (io.StringIO() or io.BytesIO() which can pass to file reader
        """
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(key)
        return blob.download_as_bytes()
    
    def save_obj(self, body: bytes, remote_path: str, bucket_name: str = None) -> 1:
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(key)
        buffer = io.BytesIO(body)
        blob.upload_from_string(data = buffer.getvalue(), content_type=None)
        return 1
    
    def save_iobuffer_to_obj(self, buffer: io.BytesIO, remote_path: str, bucket_name: str = None):
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(key)
        blob.upload_from_string(data = buffer.getvalue(), content_type=None)
        
    def read_obj_to_iobuffer(self, remote_path: str, bucket_name: str = None) -> io.BytesIO:
        obj = self.read_obj(remote_path, bucket_name)
        buffer = io.BytesIO(obj)
        return buffer
    
    def read_csv(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read CSV from GCS directly into Pandas

        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_csv(*args, **kws)
        :param kws: the keyword args passed to pd.read_csv(*args, **kws)
        :return: pandas dataframe
        """
        buffer = self.read_obj_to_iobuffer(remote_path, bucket_name)
        df = pd.read_csv(buffer, *args, **kws)
        return df
    
    def save_csv(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to csv on GCS

        :param df: pandas dataframe
        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_csv(*args, **kws)
        :param kws: the keyword args passed to pd.to_csv(*args, **kws)
        :return:
        """
        buffer = io.BytesIO()
        df.to_csv(buffer, *args, **kws)

        self.save_iobuffer_to_obj(buffer, remote_path, bucket_name)
        
    def read_parquet(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read Parquet file from GCS directly into Pandas

        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.read_parquet(*args, **kws)
        :return:
        """
        buffer = self.read_obj_to_iobuffer(remote_path, bucket_name)
        df = pd.read_parquet(buffer, *args, **kws)
        return df
    
    def save_parquet(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to parquet file on GCS

        :param df: pandas dataframe
        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.to_parquet(*args, **kws)
        :return:
        """
        buffer = io.BytesIO()
        df.to_parquet(buffer, *args, **kws)

        self.save_iobuffer_to_obj(buffer, remote_path, bucket_name)
        
    def upload_folder(self, local_folder_path: str, remote_root_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        """Upload a local folder to GCS

        :param local_folder_path: local folder path (full path)
        :param remote_root_path: root path where the folder will be created on GCS (excluding bucket path)
        :param bucket_name:
        :param multipart_threshold: an argument controlling uploading size per partition
        :return:
        """
        remote_root_path = remove_begin_slash(remote_root_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        local_parent_dir_path = os.path.dirname(local_folder_path)

        for path, subdirs, files in os.walk(local_folder_path):
            dirname = path.replace(local_parent_dir_path, '')
            for file in files:
                local_filepath = os.path.join(path, file)
                remote_filepath = os.path.join(dirname, file)
                if os.path.sep == '\\':
                    remote_filepath = remote_filepath.replace("\\", "/")
                remote_filepath = add_tail_slash(remove_begin_slash(remote_root_path)) +  remove_begin_slash(remote_filepath)
                # upload
                blob = bucket.blob(remote_filepath)
                blob.upload_from_filename(local_filepath)
        
        logging.info(f"Successfully upload folder from {local_folder_path} to {remote_root_path}")
                
    def download_folder(self, remote_folder_path: str, local_root_path: str, bucket_name: str = None):
        """

        :param remote_folder_path: folder path on S3 (excluding bucket path)
        :param local_root_path: local root path where the folder will be downloaded to (full root path)
        :param bucket_name:
        :return:
        """
        remote_folder_path = remove_begin_slash(remote_folder_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        bucket = self.client.get_bucket(bucket_name)
        
        blobs = bucket.list_blobs(prefix=remote_folder_path)
        for blob in blobs:
            remote_filepath = blob.name
            local_filepath = os.path.join(local_root_path, os.path.relpath(remote_filepath, remote_folder_path))
            if not os.path.exists(os.path.dirname(local_filepath)):
                os.makedirs(os.path.dirname(local_filepath))
            if remote_filepath.endswith("/"):  # bypass void path
                continue
            
            blob.download_to_filename(local_filepath)
            
        logging.info(f"Successfully download folder from {remote_folder_path} to {local_root_path}")


import logging
from dataclasses import dataclass, asdict
from dacite import from_dict
from collections import OrderedDict
from enum import Enum
from typing import Dict, List, Any, Optional
import ibis
from ibis import schema
from ibis.expr.schema import Schema
from ibis.expr.datatypes import DataType

def extract_dtype(dtype: str, args: dict) -> DataType:
    if dtype == 'Array':
        value_type = args.pop('value_type')
        value_dtype = extract_dtype(value_type['dtype'], value_type['args'])
        return ibis.expr.datatypes.Array(
            value_type = value_dtype,
            **args
        )
    elif dtype == 'Map':
        key_type = args.pop('key_type')
        key_dtype = extract_dtype(key_type['dtype'], key_type['args'])
        value_type = args.pop('value_type')
        value_dtype = extract_dtype(value_type['dtype'], value_type['args'])
        return ibis.expr.datatypes.Map(
            key_type = key_dtype,
            value_type = value_dtype,
            **args
        )
    elif dtype == 'Struct':
        fields = args.pop("fields")
        fields_dtype = dict()
        for k, v in fields.items():
            v_dtype = extract_dtype(v['dtype'], v['args'])
            fields_dtype[k] = v_dtype
        return ibis.expr.datatypes.Struct(
            fields = fields_dtype,
            **args
        )
    else:
        return getattr(ibis.expr.datatypes, dtype)(**args)

@dataclass
class DSchemaField:
    """data schema field structure

    :param dtype: a string representing the basic type, support Array, Map, Struct
        from from https://ibis-project.org/reference/datatypes
    :param args: the argument for each schema fields, e.g., nullable: true
    :param primary_key: whether this field is part of primary key
    :param descr: a description
    :param extra_kws: extra keyword argument may be useful
    
    example settings:
        "CUST_ID" : {
            "dtype":"Int64",
            "args":{
                "nullable": false
            },
            "descr" : "customer id",
            "primary_key" : true
        },
        "SNAP_DT" : {
            "dtype":"Date",
            "args":{
                "nullable": false
            },
            "descr" : "observation snapshot date",
            "primary_key" : true
        },
        "NAME" : {
            "dtype":"String",
            "args":{
                "nullable": false
            },
            "descr" : "first and last name",
            "primary_key" : false
        },            
        "SAMPLE_ARRAY":{
            "dtype":"Array",
            "args":{
                "value_type":{
                    "dtype":"String",
                    "args":{
                        "nullable":false
                    }
                },
                "nullable":false
            },
            "descr" : "sample array",
        },
        "SAMPLE_MAP":{
            "dtype":"Map",
            "args":{
                "key_type":{
                    "dtype":"String",
                    "args":{
                        "nullable":false
                    }
                },
                "value_type":{
                    "dtype":"Float64",
                    "args":{
                        "nullable":false
                    }
                },
                "nullable":false
            },
            "descr" : "sample map",
        },
        "SAMPLE_NESTED":{
            "dtype":"Array",
            "args":{
                "value_type":{
                    "dtype":"Map",
                    "args":{
                        "key_type":{
                            "dtype":"String",
                            "args":{
                                "nullable":false
                            }
                        },
                        "value_type":{
                            "dtype":"Int64",
                            "args":{
                                "nullable":true
                            }
                        },
                        "nullable":false
                    }
                },
                "nullable":false
            },
            "descr" : "sample nested",
        },
        "SAMPLE_STRUCT":{
            "dtype":"Struct",
            "args":{
                "fields":{
                    "id":{
                        "dtype":"Int8",
                        "args":{
                            "nullable":false
                        }
                    },
                    "trans_dt":{
                        "dtype":"Timestamp",
                        "args":{
                            "timezone":"UTC",
                            "nullable":false
                        }
                    },
                    "amount":{
                    "dtype":"Decimal",
                        "args":{
                            "precision":10,
                            "scale":2,
                            "nullable":false
                        }
                    }
                },
                "nullable":false
            },
            "descr" : "sample struct",
        }
    """
    dtype: str
    args: Dict[str, Any]
    primary_key: bool = False
    cluster_key: bool = False
    partition_key: bool = False
    descr: Optional[str] = None
    extra_kws: Optional[Dict[str, Any]] = None
    
    @property
    def ibis_dtype(self) -> DataType:
        return extract_dtype(dtype = self.dtype, args = self.args)


class DSchema(OrderedDict):
    def __init__(self, dschema_fields: Dict[str, DSchemaField]):
        """create DSchema object from predefined dschema fields

        :param dschema_fields: {feature_name, feature_type} type should use DSchemaField
        """
        super().__init__(**dschema_fields)
        
    @property
    def ibis_schema(self) -> Schema:
        sh = []
        for col, schema_field in self.items():
            sh.append(
                (col, schema_field.ibis_dtype)
            )
            
        return schema(pairs = sh)
    
    @property
    def primary_keys(self) -> List[str]:
        # return primary keys
        return [k for k, v in self.items() if v.primary_key]
    
    @property
    def cluster_keys(self) -> List[str]:
        # return primary keys
        return [k for k, v in self.items() if v.cluster_key]
    
    @property
    def partition_keys(self) -> List[str]:
        # return primary keys
        return [k for k, v in self.items() if v.partition_key]
    
    @property
    def descrs(self) -> Dict[str, str]:
        # extract any column descriptions
        return {k: v.descr for k, v in self.items() if v.descr}
            
    @classmethod
    def from_js(cls, js: dict):
        """create DSchema object from dictionary (json format)

        :param js: the dictionary of configs, example see DSchemaField doc
        :return: DSchema
        """
        s = DSchema({})
        for col, config in js.items():
            s[col] = from_dict(
                data_class = DSchemaField,
                data = config
            )
        return s
    
    def to_js(self) -> OrderedDict:
        r = OrderedDict()
        for col, schema_field in self.items():
            r[col] = asdict(schema_field)
        return r
    
    

if __name__ == '__main__':
    pass


import numpy as np
import pandas as pd
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.ensemble import StackingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone, is_classifier, is_regressor
from sklearn.preprocessing import label_binarize, LabelEncoder
from scipy.special import softmax

from ModelingTools.FeatureEngineer.transformers import WeightedAverager
from ModelingTools.utils.checks import check_params
from ModelingTools.utils.parallel import parallel_run, delayer


def target_binarize(y, classes) -> pd.DataFrame:
    y_bin = label_binarize(y, classes=classes)
    idmax = y_bin.argmax(axis=1)
    for row_idx, idmx in enumerate(idmax):
        y_bin[row_idx, idmx:] = 1
    return pd.DataFrame(y_bin, columns=classes)


class AlwaysConstEstimator(BaseEstimator):
    def __init__(self, const):
        self.const = const

    def fit(self, X, y, **params):
        return self

    def predict(self, X):
        shape = X.shape
        r = np.zeros(shape[0])
        r[:] = self.const
        return r

    def predict_proba(self, X):
        shape = X.shape
        r = np.ones(shape[0])
        return r


class OrdinalClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, estimator, ordered_classes: list = None, n_jobs: int = -1):
        """ordinal classification problem
        will decompose the problem into several binary classification problem and calculate the diff probs for each clf
        and use argmax of class prob to predict

        :param estimator:base binary classifier to use
        :param ordered_classes: specify y labels rank from least to most
        :param n_jobs:
        """
        self.estimator = estimator
        self.ordered_classes = np.asarray(ordered_classes)
        self.n_jobs = n_jobs

    @property
    def classes_(self):
        return self.ordered_classes

    @property
    def binary_estimators_(self):
        return self._bin_ests

    def fit(self, X, y, **fit_params):

        y_bins = target_binarize(y, classes=self.ordered_classes)

        self._bin_ests = {}
        # add last classifier (always 1)
        self._bin_ests[self.ordered_classes[-1]] = AlwaysConstEstimator(const=1).fit(X, y)

        # support multi-processor running
        jobs = (self._fit_one(X, y_bins.loc[:, cl], **fit_params) for cl in self.ordered_classes[:-1])
        clfs = parallel_run(jobs, n_jobs=self.n_jobs)
        self._bin_ests.update({cl: clf for cl, clf in zip(self.ordered_classes[:-1], clfs)})

        return self

    @delayer
    def _fit_one(self, X, y_bin, **fit_params):
        clf = clone(self.estimator)
        clf.fit(X, y_bin, **fit_params)
        return clf

    def predict(self, X) -> np.ndarray:
        probas = self.predict_proba(X)
        return pd.DataFrame(probas, columns=self.classes_).idxmax(axis=1).values

    @delayer
    def _predict_prob_one(self, cl, clf, X):
        prob = clf.predict_proba(X)
        if len(prob.shape) == 2:
            prob = prob[:, 1]  # binary classification, use second column
        return cl, prob  # make sure the class and tree prob is linked correctly

    def predict_proba(self, X) -> np.ndarray:
        diff_probs = self.decision_function(X)
        return softmax(diff_probs, axis=1)

    def predict_log_proba(self, X) -> np.ndarray:
        return np.log(self.predict_proba(X))

    def decision_function(self, X) -> np.ndarray:
        jobs = (self._predict_prob_one(cl, clf, X) for cl, clf in self.binary_estimators_.items())
        tuples = parallel_run(jobs, n_jobs=self.n_jobs)  # [(cl1, clf1), (cl2, clf2), (cl3, clf3), ...]
        probs = dict(tuples)

        cum_probs = pd.DataFrame(probs)[self.ordered_classes]
        diff_probs = cum_probs - cum_probs.shift(periods=1, axis=1).fillna(0)
        if isinstance(diff_probs, pd.DataFrame):
            return diff_probs.values
        else:
            return diff_probs

    def score(self, X, y, sample_weight=None) -> float:
        # returns kendal tau correlation coefficient
        from scipy import stats
        y_predict = self.predict(X)
        k, p = stats.kendalltau(y, y_predict, nan_policy='omit')
        return k


class OrdinalClassifierPlus(BaseEstimator, ClassifierMixin):
    def __init__(self, estimator, ordered_classes: list = None, output_layer: str = 'softmax', n_jobs: int = -1):
        """ordinal classification problem
        will decompose the problem into several binary classification problem and calculate the diff probs for each clf

        :param estimator:base binary classifier to use
        :param ordered_classes: specify y labels rank from least to most
        :param output_layer: {'softmax', 'lda', 'qda', 'wa+lda', 'wa+qda'}, will be applied after each clf output a prob to get the final decision (stacking)
                -- softmax: will choose the argmax prob of each class
                -- lda/qda: use lda or qda to make final decision
                -- wa+lda/wa+qda: add weighted average before feeding into lda/qda, this way will be uni-variate X and y classification
        :param n_jobs:
        """
        self.estimator = estimator
        self.ordered_classes = np.asarray(ordered_classes)
        self.output_layer = check_params(output_layer, allowed_values=['softmax', 'lda', 'qda', 'wa+lda', 'wa+qda'])
        self.n_jobs = n_jobs

        # stacking will automatically labelize y, which will conflict our base estimator, so need to manually map y to ints and map back when predict
        self._label_enc = LabelEncoder().fit(ordered_classes)
        if output_layer == 'softmax':
            final_est = LogisticRegression()
        elif output_layer == 'lda':
            final_est = LinearDiscriminantAnalysis(store_covariance=True)
        elif output_layer == 'qda':
            final_est = QuadraticDiscriminantAnalysis(store_covariance=True)
        elif output_layer == 'wa+lda':
            k = len(ordered_classes)
            final_est = Pipeline([
                ('wa', WeightedAverager(weights=range(1, k + 1), intercept=0.0, n_jobs=n_jobs)),
                ('lda', LinearDiscriminantAnalysis(store_covariance=True))]
            )
        elif output_layer == 'wa+qda':
            k = len(ordered_classes)
            final_est = Pipeline([
                ('wa', WeightedAverager(weights=range(1, k + 1), intercept=0.0, n_jobs=n_jobs)),
                ('qda', QuadraticDiscriminantAnalysis(store_covariance=True))]
            )

        self.stack_ = StackingClassifier(
            estimators=[('ordinal',
                         OrdinalClassifier(estimator, ordered_classes=self._label_enc.transform(ordered_classes),
                                           n_jobs=n_jobs))],
            final_estimator=final_est,
            stack_method='decision_function',  # use decision function of the base estimators
            n_jobs=n_jobs,
        )

    @property
    def base_ordinal_estimator_(self):
        return self.stack_.named_estimators_['ordinal']

    @property
    def final_estimator_(self):
        return self.stack_.final_estimator_

    @property
    def feature_names_in_(self):
        return self.stack_.feature_names_in_

    @property
    def classes_(self):
        return self.ordered_classes

    @property
    def binary_estimators_(self):
        return self.base_ordinal_estimator_._bin_ests

    def fit(self, X, y, **fit_params):
        y_labled = self._label_enc.transform(y)
        self.stack_.fit(X, y_labled, **fit_params)
        return self

    def predict(self, X):
        y = self.stack_.predict(X)
        # map back encoded y to original y space
        return self._label_enc.inverse_transform(y)

    def predict_proba(self, X):
        # need to follow original sequence of classes_
        p = self.stack_.predict_proba(X)
        classes_ = self._label_enc.inverse_transform(self.stack_.classes_)
        p = pd.DataFrame(p, columns=classes_)[self.classes_]
        return p.values

    def decision_function(self, X):
        # need to follow original sequence of classes_
        p = self.stack_.decision_function(X)
        classes_ = self._label_enc.inverse_transform(self.stack_.classes_)
        p = pd.DataFrame(p, columns=classes_)[self.classes_]
        return p.values

    def score(self, X, y, sample_weight=None):
        # returns kendal tau correlation coefficient
        from scipy import stats
        y_predict = self.predict(X)
        k, p = stats.kendalltau(y, y_predict, nan_policy='omit')
        return k



import logging
import pandas as pd
import numpy as np
from functools import partial
from typing import Union

class FeatureMeta:
    def __init__(self, features_meta: pd.DataFrame):
        """

        :param features_meta: pd.DataFrame([Varname, Vartype, ImputeStrategy, ImputeConst])

        Vartype: {Numerical, Binary, OrderedCateg, UnorderedCateg, IdPolicy}
        ImputeStrategy: {median, mean, most_frequent}
        ImputeConst: constant value to be imputed, e.g. 3, np.nan
        """
        self.features_meta = features_meta

    def subset(self, cols):
        self.features_meta = self.features_meta[self.features_meta['Varname'].isin(cols)]

    def getIdPolicyFeatures(self):
        # features that should not be included (typically Ids or policy reason (gender, geographic, marriage, racial, ...))
        return self.features_meta.loc[self.features_meta['Vartype'].isin(['IdPolicy']), 'Varname']

    def getNumericFeatures(self):
        return self.features_meta.loc[self.features_meta['Vartype'].isin(['Numerical']), 'Varname']

    def getCategFeatures(self):
        return self.features_meta.loc[
            self.features_meta['Vartype'].isin(['Binary', 'OrderedCateg', 'UnorderedCateg']), 'Varname']

    def getDiscreteFeatures(self):
        return self.features_meta.loc[self.features_meta['Vartype'].isin(['OrderedCateg', 'UnorderedCateg']), 'Varname']

    def getFeaturesByTypes(self, types: list):
        return self.features_meta.loc[self.features_meta['Vartype'].isin(types), 'Varname']

    def getImputeStrategyAndConst(self, col):
        return self.features_meta.loc[self.features_meta['Varname'] == col, 'ImputeStrategy'].iloc[0], \
               self.features_meta.loc[self.features_meta['Varname'] == col, 'ImputeConst'].iloc[0]

    def iterfeature(self):
        for idx, row in self.features_meta.iterrows():
            yield row['Varname'], row['Vartype'], row['ImputeStrategy'], row['ImputeConst']


from __future__ import annotations
from datetime import date
import logging
import time
from typing import List
from collections import OrderedDict
import pandas as pd
import pyspark
from CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from CommonTools.SnapStructure.tools import get_future_month_ends, get_past_month_ends
from CommonTools.utils import str2dt

""" Chain Executor with dependency check """

class _CurrentStream:
    def __init__(self, upstream):
        self.upstream = upstream

    def get_exc_dts(self, snap_dt: date):
        return [snap_dt]


class _FutureStream(_CurrentStream):
    def __init__(self, upstream, future: int, freq: str = 'M'):
        super().__init__(upstream)
        self.future = future
        self.freq = freq

    def get_exc_dts(self, snap_dt: date):
        return get_future_month_ends(snap_dt, forward=self.future, freq = self.freq)


class _PastStream(_CurrentStream):
    def __init__(self, upstream, history: int, freq: str = 'M'):
        super().__init__(upstream)
        self.history = history
        self.freq = freq

    def get_exc_dts(self, snap_dt: date):
        return get_past_month_ends(snap_dt, history=self.history, freq = self.freq)[::-1]
    

class ExecPlan:
    def __init__(self, obj, snap_dt: date):
        self.obj = obj
        self.snap_dt = snap_dt
        self.upstreams = []

    def __str__(self):
        return f"{self.obj}@{self.snap_dt} | upstreams = {len(self.upstreams)}"

    def attach_upstream(self, plan: ExecPlan):
        self.upstreams.append(plan)

    def exec_plan(self) -> dict:
        up_infos = []
        for up in self.upstreams:
            up_infos.append(up.exec_plan())

        return {
            'current': self,
            'upstream' : up_infos
        }
    
    def get_serial_steps(self) -> list:
        """get a serias of steps to execute

        :return list: _description_
        """
        r = []
        for up in self.upstreams:
            r.extend(up.get_serial_steps())
        r.append(self)
        return r
    
    def reduce_summary(self) -> OrderedDict:
        """combine similar steps (same obj and snap date) from serial plan and do summarize

        :return dict: _description_
        """
        summary = OrderedDict()
        steps = self.get_serial_steps()
        for step in steps:
            if step.obj in summary.keys():
                if step.snap_dt not in summary[step.obj]:
                    summary[step.obj].append(step.snap_dt)
            else:
                summary[step.obj] = [step.snap_dt]
        return summary



class SnapTableStreamGenerator:
    dm: SnapshotDataManagerBase = None
    upstreams: List[_CurrentStream] = []
    
    @classmethod
    def init(cls, snap_dt: date):
        """do some initialization if not exist or ready

        :param date snap_dt: _description_
        """
        pass

    @classmethod
    def execute(cls, snap_dt: date):
        """key logic for generating data at given snap date

        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")
    
    @classmethod
    def if_success(cls, snap_dt: date) -> bool:
        """determine if the run is successful or not

        :param date snap_dt: _description_
        :return bool: if True will run next job/upstream, otherwise will delete and rerun
        """
        return cls.dm.is_valid(snap_dt = snap_dt)

    @classmethod
    def _execute(cls, snap_dt: date):
        """execute the excute function defined

        :param date snap_dt: _description_
        """
        logging.info(f"Running {cls.dm}@{snap_dt}")
        snap_dt = str2dt(snap_dt)
        # single run, without checking the upstream tables

        start = time.time()
        cls.execute(snap_dt = snap_dt)
        end = time.time()
        seconds = end - start
        logging.info(f"Saved to {cls.dm}@{snap_dt}, used time {seconds // 60 : .0f} minutes {seconds % 60 : .0f} seconds")

    
    @classmethod
    def _prerun(cls, snap_dt: date):
        """preparation before actual run
        

        :param date snap_dt: _description_
        :return _type_: _description_
        """
        if cls.dm.exist():
            if snap_dt in cls.dm.get_existing_snap_dts():
                if cls.if_success(snap_dt):
                    logging.info(f"Already extracted for {cls.dm}@{snap_dt}, will bypass")
                    return
                else:
                    logging.warning(f"Existing records not valid for {cls.dm}@{snap_dt}, will rerun")
                    cls.dm.delete(snap_dt)
        else:
            logging.info(f"Table-Schema does not exist, will initialize")
            cls.init(snap_dt)
    
    @classmethod
    def run(cls, snap_dt: date):
        """run current and upstreams

        :param date snap_dt: _description_
        """
        snap_dt = str2dt(snap_dt)
        # prepare
        cls._prerun(snap_dt)

        # check upstream table recursively and make up any missing files
        for up in cls.upstreams:
            logging.info(f"Upstream found: {up.upstream.dm}")
            for dt in up.get_exc_dts(snap_dt):
                logging.info(f"Run upstream {up.upstream.dm} @ {dt}")
                up.upstream.run(dt)

        # execute
        cls._execute(snap_dt)
        
    @classmethod
    def run_detached(cls, snap_dt: date):
        """run current timestamp only, ignore upstreams

        :param date snap_dt: _description_
        :return _type_: _description_
        """
        snap_dt = str2dt(snap_dt)
        # prepare
        cls._prerun(snap_dt)
            
        # execute
        cls._execute(snap_dt)
            

    @classmethod
    def get_exec_plan(cls, snap_dt: date) -> ExecPlan:
        """get the execution plan object for further analysis

        :param date snap_dt: _description_
        :return ExecPlan: _description_
        """
        snap_dt = str2dt(snap_dt)

        plan = ExecPlan(
            obj = cls,
            snap_dt = snap_dt
        )

        for up in cls.upstreams:
            for dt in up.get_exc_dts(snap_dt):
                plan.attach_upstream(up.upstream.get_exec_plan(dt))

        return plan



class SETTINGS:
    DB_CREDENTIAL_PATH = None
    DB_CREDENTIAL_INTERACTIVE = False

    HTTP_PROXY = None
    HTTPS_PROXY = None

    @ classmethod
    def SET_CREDENTIAL_INTERACTIVE(cls):
        """set to interactive mode, good for JupyterHub

        :return:
        """
        cls.DB_CREDENTIAL_INTERACTIVE = True
        cls.DB_CREDENTIAL_PATH = None

    @ classmethod
    def SET_CREDENTIAL_PATH(cls, path = None):
        """set to script mode, and specify a credential ini file path

        :param path: should be an .ini or .toml configuration file following the below format:
        :return:

        example credential.ini:
        [Database]
        username = xxxxxxx
        password = xxxxxxx
        """
        cls.DB_CREDENTIAL_INTERACTIVE = False
        cls.DB_CREDENTIAL_PATH = path

    @ classmethod
    def SET_PROXY_SERVER(cls, http_proxy: str, https_proxy: str):
        cls.HTTP_PROXY = http_proxy
        cls.HTTPS_PROXY = https_proxy


from datetime import date, timedelta
import glob
import os
from typing import List
import re
import pandas as pd
from CommonTools.utils import get_month_end_date

def get_forward_month_end_date(dt: date) -> date:
    next_mth_start = get_month_end_date(dt.year, dt.month) + timedelta(days=1)
    return get_month_end_date(next_mth_start.year, next_mth_start.month)

def get_last_month_end_date(dt: date) -> date:
    return dt.replace(day=1) - timedelta(days=1)

def get_file_list_pattern(folder_path: str, pattern: str, include_dir: bool = True) -> list:
    """return a list of filepath that matches the pattern

    :param folder_path:
    :param pattern:
    :param include_dir: whether to include the dirname in the result
    :return: a list of macthed file paths (including/excluding directory path)

    >>> get_file_list_pattern("/app/CMM_BFS_CRI_BRR/Model_Dev_2022/Model_Dev_Pipeline/1_raw_data_extraction", pattern = "*clnt*", include_dir = False)
    """
    files_with_dir = glob.glob(f"{folder_path}/{pattern}")
    if include_dir:
        return files_with_dir
    else:
        return [os.path.basename(f) for f in files_with_dir]


def match_date_from_str(s:str) -> str:
    """

    :param s: string, date should be in YYYY-MM-DD format or YYYY-DD-MM format
    :return:
    """
    m = re.search("([0-9]{4}-[0-9]{2}-[0-9]{2})", s)
    if m:
        return m.group()
    else:
        return

def generate_cohort_list(start_month: date, end_month: date, freq: str = 'M') -> List[date]:
    """

    :param start_month: start date of the cohort
    :param end_month: end date of the cohort
    :param freq: # https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases
    :return:

    https://pandas.pydata.org/docs/reference/api/pandas.date_range.html
    """
    ts = pd.date_range(start = start_month, end = end_month, freq = freq).to_pydatetime()
    return [t.date() for t in ts]

def get_past_month_ends(snap_dt: date, history:int = 12, freq: str = 'M') -> List[date]:
    # return a list of month end date for the past 12 months prior to the snap_dt
    ts = pd.date_range(end = snap_dt, freq = freq, periods = history).to_pydatetime()
    return [t.date() for t in ts]

def get_future_month_ends(snap_dt: date, forward:int = 12, freq: str = 'M') -> List[date]:
    # return a list of month end date for the past 12 months prior to the snap_dt
    ts = pd.date_range(start = snap_dt, freq = freq, periods = forward).to_pydatetime()
    return [t.date() for t in ts]


from pathlib import Path
import os
import yaml
from botocore.exceptions import ClientError
from CommonTools.dtyper import DSchema
from CommonTools.obj_storage import ObjStorage


class BaseSchemaManager:
    def write(self, schema: str, table: str, dschema: DSchema):
        content = dschema.to_js()
        self.write_raw(schema, table, content)
    
    def write_raw(self, schema: str, table: str, content: dict):
        raise NotImplementedError("")
    
    def read(self, schema: str, table: str) -> DSchema:
        content = self.read_raw(schema, table)
        return DSchema.from_js(content)
    
    def read_raw(self, schema: str, table: str) -> dict:
        raise NotImplementedError("")
    

class LocalSchemaManager(BaseSchemaManager):
    """Table Schema Manager on local File System
    """
    def __init__(self, root_dir: str):
        self.root_dir = root_dir
        
    def write_raw(self, schema: str, table: str, content: dict):
        # create path
        schema_folder = Path(self.root_dir) / schema
        schema_folder.mkdir(parents = True, exist_ok = True)
        # save file
        schema_file = schema_folder / f"{table}.yml"
        with open(schema_file, 'w') as obj:
            yaml.dump(content, obj, default_flow_style=False, sort_keys=False)
            
    def read_raw(self, schema: str, table: str) -> dict:
        schema_file = Path(self.root_dir) / schema / f"{table}.yml"
        try:
            with open(schema_file, 'r') as obj:
                record = yaml.safe_load(obj)
        except FileNotFoundError as e:
            raise ValueError("No record found for given schema and table")
        else:
            return record
        
        
class ObjStorageSchemaManager(BaseSchemaManager):
    """Table Schema Manager on object storage (S3, GCS, etc.)
    """
    def __init__(self, obj_st: ObjStorage, bucket: str, root_dir: str):
        obj_st.enter_bucket(bucket)
        self._obj_st = obj_st
        self.root_dir = root_dir
        
    def write_raw(self, schema: str, table: str, content: dict):
        # create content
        content_str = yaml.dump(content, default_flow_style=False, sort_keys=False)
        content_bytes = bytes(content_str, encoding = 'utf8')
        # save config
        schema_file = os.path.join(self.root_dir, schema, f"{table}.yml")
        self._obj_st.save_obj(
            body = content_bytes,
            remote_path = schema_file
        )
        
    def read_raw(self, schema: str, table: str) -> dict:
        schema_file = os.path.join(self.root_dir, schema, f"{table}.yml")
        try:
            obj = self._obj_st.read_obj(remote_path = schema_file)
            record = yaml.safe_load(obj)
        except ClientError as e:
            raise ValueError("No record found for given schema and table")
        return record
    


from typing import Union, Dict, List, Literal
import pandas as pd
import numpy as np
from scipy.stats import chi2_contingency
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler, RobustScaler
from sklearn.feature_selection import mutual_info_classif
from sklearn.impute import SimpleImputer
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from category_encoders.ordinal import OrdinalEncoder
from category_encoders.count import CountEncoder
from ModelingTools.CustomModel.linear import LinearClfStatsModelWrapper
    
def chi2_score_cross(X: pd.DataFrame) -> pd.DataFrame:
    categ_cols = X.columns
    result = []
    for c1 in categ_cols:
        for c2 in categ_cols:
            #if  c1 != c2:
            chi2, pvalue, dof, matrix = chi2_contingency(
                pd.crosstab(
                    X[c1],
                    X[c2]
                )
            )
            result.append((c1, c2, chi2, pvalue))
    chi_test_output = pd.DataFrame(
        result, 
        columns = ['var1', 'var2', 'chi2', 'pvalue']
    )
    return chi_test_output

def chi2_score_matrix(X: pd.DataFrame) -> pd.DataFrame:
    chi_test_output = chi2_score_cross(X)
    grid = chi_test_output.pivot(
        index='var1', 
        columns='var2', 
        values='pvalue'
    )
    grid.index.name = 'Feature'
    grid.columns.name = 'Features'
    return grid

def get_corr_matrix(df: pd.DataFrame) -> pd.DataFrame:
    grid = df.corr()
    grid.index.name = 'Feature'
    grid.columns.name = 'Features'
    return grid


import pandas as pd
import numpy as np
from scipy.stats import ks_2samp
from sklearn.metrics._ranking import _binary_clf_curve
from sklearn.metrics import make_scorer, get_scorer, log_loss

def ks_stat(y, yhat):
    return ks_2samp(yhat[y==1], yhat[y!=1]).statistic

def binary_clf_metric_at_thresholds(y_true, pred_prob) -> pd.DataFrame:
    """Binary classification metrics at each threshold

    :param y_true: ground truth target array (binary)
    :param pred_prob: predicted probabilities array (float between 0-1)
    :return: DataFrame of columns [threshold, tp, fp, tn, fn, tpr, fpr, precision, recall]
    """
    fps, tps, thresholds = _binary_clf_curve(y_true, pred_prob)
    num_p = y_true.sum()
    num_n = (1 - y_true).sum()

    t = pd.DataFrame({
        'threshold': thresholds,
        'tp': tps,
        'fp': fps,
        'tn': num_n - fps,
        'fn': num_p - tps
    })
    t['tpr'] = t['tp'] / (t['tp'] + t['fn'])
    t['fpr'] = t['fp'] / (t['fp'] + t['tn'])
    t['precision'] = t['tp'] / (t['tp'] + t['fp'])
    t['recall'] = t['tpr']
    return t


import io
from typing import List, Any
import pandas as pd

def remove_begin_slash(url: str) -> str:
    if url.startswith('/'):
        prefix = str(url)[1:]
    else:
        prefix = str(url)
    return prefix

def add_tail_slash(url: str) -> str:
    if url.endswith('/'):
        return str(url)
    else:
        return str(url) + '/'

class ObjStorage:
    
    def get_uri_protocol(self) -> str:
        raise NotImplementedError("")

    def list_buckets(self) -> pd.DataFrame:
        """return bucket list [Name, CreationDate]
        """
        raise NotImplementedError("")

    def create_bucket(self, bucket_name: str) -> int:
        """Create a new bucket

        :param bucket_name: name of the bucket
        :return: 1 if success
        """
        raise NotImplementedError("")

    def get_bucket(self, bucket_name: str = None):
        raise NotImplementedError("")


    def enter_bucket(self, bucket_name: str = None) -> int:
        """move to a given bucket

        :param bucket_name: name of the bucket you want to switch to
        :return: 1 if success
        """
        raise NotImplementedError("")

    def get_current_bucket(self):
        """get your current bucket object

        :return:
        """
        raise NotImplementedError("")

    def delete_bucket(self, bucket_name: str, interactive: bool = True) -> int:
        """delete the bucket

        :param bucket_name: name of the bucket to delete
        :param interactive: whether you would like an interactive warning message
        :return: 1 if success
        """
        raise NotImplementedError("")


    ''' below requires lock a bucket '''

    def list_objs(self, remote_path: str, bucket_name: str = None) -> pd.DataFrame:
        """list objects under the specific directory

        :param remote_path: remote path excluding bucket
        :param bucket_name: name of the bucket, if not given, will use the default bucket
        :return: pandas dataframe showing files inside ['Key', 'LastModified', 'Size']
        """
        raise NotImplementedError("")

    def get_obj(self, remote_path:str, bucket_name: str = None) -> Any:
        raise NotImplementedError("")

    def create_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Create an object

        :param remote_path: if remote path ends with /, it will be folder, otherwise will be a blank file
        :param bucket_name:
        :return: 1 if success otherwise the info dictionary
        """
        raise NotImplementedError("")

    def delete_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Delete an object/file

        :param remote_path: the filepath to delete (excluding bucket path)
        :param bucket_name: in which bucket
        :return:
        """
        raise NotImplementedError("")

    def create_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Create a folder object on object storage

        :param remote_folder_path: the folder path (excluding bucket path)
        :param bucket_name: under which bucket
        :return:
        """
        raise NotImplementedError("")

    def delete_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Delete the folder and every file underneath it

        :param remote_folder_path: the folder path to delete (excluding bucket path)
        :param bucket_name:
        :return:
        """
        raise NotImplementedError("")

    def copy_obj(self, from_bucket_name: str = None, from_path: str = None, to_bucket_name: str = None, to_path: str = None):
        """Copy one file from one place to another (can copy to a different bucket)

        :param from_bucket_name: from which bucket you want to copy the file
        :param from_path: which file you want to copy (excluding bucket path)
        :param to_bucket_name: to which bucket you want to copy to
        :param to_path: the destination file path (excluding bucket path)
        :return:
        """
        raise NotImplementedError("")


    def upload_file(self, local_path: str, remote_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        '''upload file from local machine to object storage

        :param local_path: local file path that you would like to upload to S3, suggest to be absolute path
        :param remote_path: remote path on MinIO (excluding bucket path); could either start with / or not
        :param multipart_threshold: Setting an extremely large multipart threshold to effectively disable multipart uplaod.
                Note: Multiplart upload is not compatible with GCS, therefore disabling multipart upload for compatability with hybrid solution.
                set it to None is you want multipart saving
        :return:
        '''
        raise NotImplementedError("")

    def download_file(self, remote_path: str, local_path: str, bucket_name: str = None):
        '''download file from object storage to local machine

        :param remote_path: remote path on object storage (excluding bucket); could either start with / or not
        :param local_path: local file path that you would like to save the file from MinIO, suggest to be absolute path
        :param bucket_name
        :return:
        '''
        raise NotImplementedError("")


    def read_obj(self, remote_path: str, bucket_name: str = None) -> bytes:
        """read obj from object storage without downloading

        :param remote_path:  remote path to read on object storage (excluding bucket); could either start with / or not
        :param Bucket:
        :return: a buffer (io.StringIO() or io.BytesIO() which can pass to file reader
        """
        raise NotImplementedError("")

    def save_obj(self, body: bytes, remote_path: str, bucket_name: str = None) -> 1:
        raise NotImplementedError("")

    def read_obj_to_iobuffer(self, remote_path: str, bucket_name: str = None) -> io.BytesIO:
        raise NotImplementedError("")

    def save_iobuffer_to_obj(self, buffer: io.BytesIO, remote_path: str, bucket_name: str = None):
        raise NotImplementedError("")

    def read_csv(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read CSV from S3 directly into Pandas

        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_csv(*args, **kws)
        :param kws: the keyword args passed to pd.read_csv(*args, **kws)
        :return: pandas dataframe
        """
        raise NotImplementedError("")

    def save_csv(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to csv on object storage

        :param df: pandas dataframe
        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_csv(*args, **kws)
        :param kws: the keyword args passed to pd.to_csv(*args, **kws)
        :return:
        """
        raise NotImplementedError("")

    def read_parquet(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read Parquet file from object storage directly into Pandas

        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.read_parquet(*args, **kws)
        :return:
        """
        raise NotImplementedError("")

    def save_parquet(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to parquet file on object storage

        :param df: pandas dataframe
        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.to_parquet(*args, **kws)
        :return:
        """
        raise NotImplementedError("")

    def upload_folder(self, local_folder_path: str, remote_root_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        """Upload a local folder to object storage

        :param local_folder_path: local folder path (full path)
        :param remote_root_path: root path where the folder will be created on S3 (excluding bucket path)
        :param bucket_name:
        :param multipart_threshold: an argument controlling uploading size per partition
        :return:
        """
        raise NotImplementedError("")


    def download_folder(self, remote_folder_path: str, local_root_path: str, bucket_name: str = None):
        """

        :param remote_folder_path: folder path on object storage (excluding bucket path)
        :param local_root_path: local root path where the folder will be downloaded to (full root path)
        :param bucket_name:
        :return:
        """
        raise NotImplementedError("")



import logging
import os
import socket
from pyspark.sql import SparkSession

from CommonTools.data_structure import EnhancedDict
from CommonTools.dbapi import baseDbInf


class SparkConnector:

    @classmethod
    def setup(cls, jdbc_jars: list = None, s3_config : dict = None, **additional_params: dict):
        """set up for the environment to run successfully

        :param jdbc_jars: if you want to connect to database (MySQL, SQL server, Hive, Teradata, etc) set this. ['path_to_jar1', 'path_to_jar2]
        :param s3_config: if you want to connect to AWS S3, set s3_configuration to connect, need to provide: {'STORAGE_URL', 'ACCESS_KEY', 'SECRET_ACCESS_KEY', 'SESSION_TOKEN'}
        :param additional_params: additional kws to pass to spark config
        :return:
        """

        cls.JAR_PATHS = ":".join(jdbc_jars)
        logging.info(f"JDBC Driver Jars: {cls.JAR_PATHS}")

        ## build the configs
        cls.conf = EnhancedDict()
        cls.conf["spark.sql.execution.arrow.pyspark.enabled"] = True  # fast transform from pandas to spark
        cls.conf["spark.executor.allowSparkContext"] = True # https://learn.microsoft.com/en-us/answers/questions/941566/sparkcontext-should-only-be-created-and-accessed-o.html

        if s3_config:
            cls.conf["spark.hadoop.fs.s3a.endpoint"] = s3_config.get("STORAGE_URL")
            cls.conf["spark.hadoop.fs.s3a.aws.credentials.provider"] = 'org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider'
            cls.conf["spark.hadoop.fs.s3a.access.key"] = s3_config.get("ACCESS_KEY")
            cls.conf["spark.hadoop.fs.s3a.secret.key"] = s3_config.get("SECRET_ACCESS_KEY")
            cls.conf["spark.hadoop.fs.s3a.session.token"] = s3_config.get("SESSION_TOKEN")
            cls.conf["spark.hadoop.fs.s3a.fast.upload"] = True
            cls.conf["spark.hadoop.fs.s3a.path.style.access"] = True
            cls.conf["spark.hadoop.fs.s3a.impl"] = "org.apache.hadoop.fs.s3a.S3AFileSystem"
            cls.conf["spark.hadoop.fs.s3a.multipart.size"] = "128M"
            cls.conf["spark.hadoop.fs.s3a.fast.upload.active.blocks"] = 8
        if jdbc_jars:
            cls.conf["spark.driver.extraClassPath"] = cls.JAR_PATHS

        for k, v in additional_params.items():
            cls.conf[k] = v


    @classmethod
    def getSparkSession(cls, app_name: str = "SparkConnector", master: str = "local[*]"):

        sparkSessionBuilder = SparkSession.builder.master(master).appName(app_name)

        for k, v in cls.conf.items():
            sparkSessionBuilder = sparkSessionBuilder.config(k, v)

        spark = sparkSessionBuilder.getOrCreate()

        return spark

    def __init__(self, app_name: str = "SparkConnector", master: str = "local"):
        if not hasattr(self, 'JAR_PATHS'):
            raise Exception("Must call SparkConnector.setup() before initialization")

        self.spark = self.getSparkSession(app_name, master)
        logging.info(
            f"Spark session object successfully created, call SparkConnector({app_name}).spark to get spark obj")

    def get_file_path(self, path: str) -> str:
        return path

    def get_common_reader(self, header=True, sep=",", schema=None):
        if schema:
            return self.spark.read.option('header', header) \
                .option("inferSchema", "false") \
                .option("sep", sep) \
                .option("encoding", "utf-8") \
                .schema(schema)
        else:
            return self.spark.read.option('header', header) \
                .option("inferSchema", "true") \
                .option("sep", sep) \
                .option("encoding", "utf-8")

    def read_csv(self, *path: str, header=True, sep=",", schema=None):
        """read csv file(s) into spark

        :param path: can either be one single csv file, or multiple files as positional arguments
        :param header: whether it has a header
        :param sep: what separator to use
        :param schema: if you have a spark schema, attach it
        :return: spark dataframe

        >>> # for the `path` argument, can read one single file as:
        >>> df = sc.read_csv(path, sep = ',')
        >>> # or read multiple csv files with same schema (usually in one folder)
        >>> df = sc.read_csv(path1, path2, path3, sep = ',')
        """
        # path = self.get_file_path(path)

        try:
            df = self.get_common_reader(header, sep, schema).csv(*path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def read_sas(self, path: str):
        # path = self.get_file_path(path)

        try:
            df = self.spark.read.format("com.github.saurfang.sas.spark").load(path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def read_parquet(self, *path: str, header=True, sep=",", schema=None):
        """read parquet file(s) into spark

        :param path: can either be one single csv file, or multiple files as positional arguments
        :param header: whether it has a header
        :param sep: what separator to use
        :param schema: if you have a spark schema, attach it
        :return: spark dataframe

        >>> # for the `path` argument, can read one single file as:
        >>> df = sc.read_parquet(path, sep = ',')
        >>> # or read multiple csv files with same schema (usually in one folder)
        >>> df = sc.read_parquet(path1, path2, path3, sep = ',')
        """
        # path = self.get_file_path(path)

        try:
            df = self.get_common_reader(header, sep, schema).parquet(*path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def save_csv(self, df, path: str, header=True, sep=",", mode="overwrite", repartition: int = None):
        """save spark dataframe to csv file (partitions saved in sub folder)

        :param df: spark dataframe
        :param path: path of the csv file (actually a folder named xxx.csv with partitions inside)
        :param header: whether to include a header
        :param sep:
        :param mode: {overwrite or append or ignore}
        :param repartition: number of partitions to split the dataframe into
        :return:
        """
        path = self.get_file_path(path)

        if repartition:
            df = df.repartition(repartition)

        try:
            df.write.option('header', header).option("sep", sep).mode(mode).csv(path)
        except Exception as e:
            logging.error('File Wrote FAILED: ' + str(e))
        else:
            logging.info('File Wrote successfully.')

    def save_parquet(self, df, path: str, header=True, sep=",", mode="overwrite", repartition: int = None):
        """save spark dataframe to parquet file (partitions saved in sub folder)

        :param df: spark dataframe
        :param path: path of the parquet file (actually a folder named xxx.csv with partitions inside)
        :param header: whether to include a header
        :param sep:
        :param mode: {overwrite or append or ignore}
        :param repartition: number of partitions to split the dataframe into
        :return:
        """
        path = self.get_file_path(path)

        if repartition:
            df = df.repartition(repartition)

        try:
            df.write.option('header', header).option("sep", sep).mode(mode).parquet(path)
        except Exception as e:
            logging.error('File Wrote FAILED: ' + str(e))
        else:
            logging.info('File Wrote successfully.')

    def query_db(self, dbConf: baseDbInf, sql: str, result_name: str = "temp_df", **kws):
        """query_extract from database and return spark dataframe

        :param dbConf: the dbConf object for connection string setup
        :param sql: the sql query_extract to make
        :param result_name: spark jdbc requires each query_extract result to have a table name
        :param kws: other parameters that could parse into .jdbc(properties)
        :return: spark DataFrame
        """

        sql = f"({sql.replace(';', '')}) {result_name}"  # drop ; and wrap with temp table name
        properties = {
            "driver": dbConf.getDriverClass(),
            # "dbtable" : sql,
        }

        if hasattr(dbConf, 'username'):
            properties["user"] = dbConf.username
            properties["password"] = dbConf.password

        for k, v in kws.items():
            properties[k] = v

        try:
            df = self.spark.read.jdbc(
                url=dbConf.getJdbcUrl(),
                table=sql,
                properties=properties
            )
        except Exception as e:
            logging.error('Query Load FAILED: ' + str(e))
        else:
            logging.info('Query Load successfully.')
            return df

    def save_db(self, dbConf: baseDbInf, df, table_name: str, mode: str = None, column_schema: dict = None,
                **kws) -> int:
        """save df to database

        :param dbConf: the dbConf object for connection string setup
        :param df: spark dataframe to save
        :param table_name: name of the table in the database
        :param mode: {None, append, overwrite}, None will fail if table exists, append will append to end, overwrite will discard existing values
        :param column_schema: dictionary of column and dtype, e.g. column_schema = {"cur" : "varchar(4)", "v1": "float"}
        :param kws: other parameters that could parse into .jdbc(properties)
        :return: 1 for success and 0 for failed
        """
        properties = {
            "driver": dbConf.getDriverClass(),
            "dbtable": table_name,
        }

        if hasattr(dbConf, 'username'):
            properties["user"] = dbConf.username
            properties["password"] = dbConf.password

        if column_schema:
            col_sch = ",".join(f"{k} {v}" for k, v in column_schema.items())
            properties["createTableColumnTypes"] = col_sch

        for k, v in kws.items():
            properties[k] = v

        try:
            df.write.jdbc(
                url=dbConf.getJdbcUrl(),
                table=table_name,
                mode=mode,
                properties=properties
            )
        except Exception as e:
            logging.error('Save to DB FAILED: ' + str(e))
            return 0
        else:
            logging.info('Save to DB successfully.')
            return 1


class SparkConnectorK8s(SparkConnector):
    @classmethod
    def setup(cls, jdbc_jars: list = None, s3_config : dict = None):
        """set up for the environment to urn successfully

        :param jdbc_jars: if you want to connect to database (MySQL, SQL server, Hive, Teradata, etc) set this. ['path_to_jar1', 'path_to_jar2]
        :param s3_config: if you want to connect to AWS S3, set s3_configuration to connect, need to provide: {'STORAGE_URL', 'ACCESS_KEY', 'SECRET_ACCESS_KEY', 'SESSION_TOKEN'}
        :return:
        """

        cls.JAR_PATHS = ":".join(jdbc_jars)
        logging.info(f"JDBC Driver Jars: {cls.JAR_PATHS}")

        os.environ['SPARK_CONF_DIR'] = os.environ["EDL_CONF_LOCATION"]
        krb_conf = os.getenv('KRB5_CONFIG')
        os.environ["SPARK_SUBMIT_OPTS"] = f"-Djava.security.krb5.conf={krb_conf}"
        SID = os.getenv("JUPYTERHUB_USER")

        os.environ['PYSPARK_PYTHON'] = 'python3.8'
        os.environ['PYSPARK_DRIVER_PYTHON'] = 'python3.8'

        cls.conf = EnhancedDict()

        cls.conf["spark.kubernetes.container.image"] = "image-path"
        cls.conf["spark.kubernetes.authenticate.caCertFile"] = "/var/run/secrets/kubernetes.io/serviceaccount/ca.crt"
        cls.conf["spark.kubernetes.authenticate.oauthTokenFile"] = "/var/run/secrets/kubernetes.io/serviceaccount/token"
        cls.conf["spark.executor.instances"] = "2"
        cls.conf["spark.executor.cores"] = "2"
        cls.conf["spark.kubernetes.executor.request.cores"] = "2"
        cls.conf["spark.kubernetes.executor.limit.cores"] = "2"
        cls.conf["spark.driver.memory"] = "4G"
        cls.conf["spark.driver.cores"] = "2"
        cls.conf["spark.executor.memory"] = "4G"

        cls.conf["spark.driver.host"] = socket.gethostbyname(socket.gethostname())
        cls.conf["spark.kubernetes.namespace"] = os.getenv("POD_NAMESPACE")
        cls.conf["spark.kubernetes.container.image.pullSecrets"] = "artifactory-regcred"
        cls.conf["spark.kubernetes.container.image.pullPolicy"] = "IfNotPresent"
        cls.conf["spark.submit.deployMode"] = "client"
        cls.conf["spark.kubernetes.authenticate.driver.serviceAccountName"] = "spark-user"
        cls.conf["spark.kubernetes.executor.podNamePrefix"] = SID
        ######
        cls.conf["spark.hadoop.fs.s3a.endpoint"] = s3_config.get("STORAGE_URL")
        cls.conf[
            "spark.hadoop.fs.s3a.aws.credentials.provider"] = 'org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider'
        cls.conf["spark.hadoop.fs.s3a.access.key"] = s3_config.get("ACCESS_KEY")
        cls.conf["spark.hadoop.fs.s3a.secret.key"] = s3_config.get("SECRET_ACCESS_KEY")
        cls.conf["spark.hadoop.fs.s3a.session.token"] = s3_config.get("SESSION_TOKEN")
        cls.conf["spark.hadoop.fs.s3a.fast.upload"] = True
        cls.conf["spark.hadoop.fs.s3a.path.style.access"] = True
        cls.conf["spark.hadoop.fs.s3a.impl"] = "org.apache.hadoop.fs.s3a.S3AFileSystem"
        cls.conf["spark.hadoop.fs.s3a.multipart.size"] = "128M"
        cls.conf["spark.hadoop.fs.s3a.fast.upload.active.blocks"] = 8
        cls.conf["spark.driver.extraClassPath"] = cls.JAR_PATHS

        ######################
        cls.conf["spark.kerberos.principal"] = f"{SID}@server-path"
        cls.conf["spark.kerberos.keytab"] = "/home/jovyan/.keytabs/edl.keytab"
        cls.conf["spark.hadoop.javax.jdo.option.ConnectionUserName"] = SID
        cls.conf["spark.hadoop.dfs.namenode.kerberos.principal.pattern"] = "*"
        cls.conf["spark.sql.hive.metastore.version"] = "1.2.1"
        cls.conf["spark.sql.hive.metastore.jars"] = "local:///" + os.environ["EDL_JARS_LOCATION"]
        cls.conf["spark.sql.execution.arrow.pyspark.enabled"] = "true"

        # cls.conf["spark.driver.extraClassPath"] = cls.JAR_PATHS
        # cls.conf["spark.executor.extraClassPath"] = os.environ["EDL_JARS_LOCATION"]
        # cls.conf["spark.executor.extraClassPath"] = os.environ["EDL_CONF_LOCATION"] + "*"

    @classmethod
    def getSparkSession(cls, app_name: str = "SparkConnectorEDL", master: str = None):
        master = "k8s://https://kubernetes.default.svc.cluster.local:443"
        sparkSessionBuilder = SparkSession.builder.master(master).appName(app_name)

        for k, v in cls.conf.items():
            sparkSessionBuilder = sparkSessionBuilder.config(k, v)

        spark = sparkSessionBuilder.enableHiveSupport().getOrCreate()

        return spark


from functools import partial
import pandas as pd

def _select_present_subset(selected_columns, df: pd.DataFrame):
    return [col for col in df.columns if col in selected_columns]

def make_present_col_selector(selected_columns):
    """ColumnTransformer support handling of missing columns

    :param selected_columns:
    :return: a column selector that can be fitted to ColumnTransformer

    Reference:
    https://github.com/scikit-learn/scikit-learn/issues/19014
    https://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html#sklearn.compose.ColumnTransformer

    e.g.
    preprocessor_new = ColumnTransformer(
    transformers=[
        ('num', numeric_transformer, make_present_col_selector(numeric_features)),
        ('cat', categorical_transformer, make_present_col_selector(categorical_features))
    ]
)
    """
    return partial(_select_present_subset, selected_columns)


def _select_present_sibling_subset(selected_columns, df: pd.DataFrame):
    # find the columns from df that are similar (siblings) of selected columns
    # sometimes column changes as result of onehot encoding and imputer indicator
    return df.columns[df.columns.str.contains('|'.join(selected_columns))] # use a regrex to define

def make_sibling_col_selector(selected_columns):
    return partial(_select_present_sibling_subset, selected_columns)


import pandas as pd
import numpy as np
from typing import Union, Dict, List, Literal
from collections import namedtuple
from scipy.special import exp10
from numpy import expm1, log1p, log10
from scipy.stats import rv_histogram, chi2_contingency, variation, kstest
from optbinning import MulticlassOptimalBinning

from ModelingTools.utils.parallel import parallel_run, delayer
from CommonTools.accessor import toJSON, loadJSON

# named tuple for value and percentage
StatVar = namedtuple(
    'StatVar', 
    field_names=['value', 'perc'], 
    defaults = [0, 0.0]
)

# named tuple for quantile stats
QuantileStat = namedtuple(
    'QuantileStat',
    field_names=['minimum','perc_1th',  'perc_5th', 'q1', 'median', 'q3', 'perc_95th', 'perc_99th', 'maximum', 'iqr', 'range']
)
# named typle for descriptive stats
DescStat = namedtuple(
    'DescStat',
    field_names=['mean', 'var', 'std', 'skew', 'kurt', 'mad', 'cv', 'normality_p']
)
# named tuple for extreme value analysis
XtremeStat = namedtuple(
    'XtremeStat',
    field_names=['lbound', 'rbound', 'lxtreme_mean', 'lxtreme_median', 'rxtreme_mean', 'rxtreme_median']
)


def serialize(v):
    if pd.isnull(v):
        return None
    if v == np.inf:
        return float('inf')
    if v == -np.inf:
        return -float('inf')
    if hasattr(v, 'item'):
        return v.item()
    else:
        return v 

class _BaseStat:
    def __init__(self):
        self.classname = self.__class__.__name__
    
    def fit(self, vector: pd.Series):
        raise NotImplemented("")
    
    def collect_attrs(self) -> dict:
        # gather serializable parameters for __init__
        raise NotImplemented("")

    def collect_params(self) -> dict:
        # gather serializable estimated parameters
        raise NotImplemented("")

    @classmethod
    def set_params(cls, obj, param_dict: dict):
        # set params estimated to the new cls object
        raise NotImplemented("")

    def to_dict(self) -> dict:
        if self.fitted_:
            return dict(
                colname = self.colname_,
                attr = self.collect_attrs(),
                params = self.collect_params()
            )
        else:
            raise Exception("Not fitted error")

    @classmethod
    def from_dict(cls, d: dict):
        c = cls(
            **d['attr']
        )
        c.colname_ = d['colname']
        cls.set_params(c, d['params'])
        c.fitted_ = True
        return c

    def generate(self, size: int = 100, seed: int = None, ignore_na:bool = False) -> pd.Series:
        raise NotImplemented("")

class CategStat(_BaseStat):
    def __init__(self, int_dtype: bool = False, max_categories: int = 25):
        super().__init__()
        self.fitted_ = False
        self.int_dtype_ = int_dtype
        self.max_categories_ = max_categories

    def fit(self, vector: pd.Series):
        self.colname_ = vector.name
        # total number of records
        self.total_ = len(vector)
        # missing rate
        missing = vector.isna()
        self.missing_ = StatVar(
            value=serialize(missing.sum()), 
            perc=serialize(missing.mean())
        )
        # num of unique value (dropped NA value)
        unique = len(vector.dropna().unique())
        self.unique_ = StatVar(
            value=serialize(unique), 
            perc=serialize(unique / self.total_)
        )
        # category counts without null
        self.vcounts_ = vector.value_counts(
            dropna = True,
            normalize = False
        ).fillna(0)
        self.vpercs_ = (self.vcounts_ / self.vcounts_.sum()).fillna(0)

        self.fitted_ = True

    def collect_attrs(self) -> dict:
        # gather serializable parameters for __init__
        return dict(
            int_dtype = self.int_dtype_,
            max_categories = self.max_categories_
        )

    def collect_params(self) -> dict:
        # gather serializable estimated parameters
        return dict(
            stats = dict(
                total = self.total_,
                missing = self.missing_._asdict(),
                unique = self.unique_._asdict()
            ),
            distribution = dict(
                vcounts = self.vcounts_.to_dict(),
                vpercs = self.vpercs_.to_dict(),
            )
        )

    @classmethod
    def set_params(cls, obj, param_dict: dict):
        # set params estimated to the new cls object
        # stats
        obj.total_ = param_dict['stats']['total']
        obj.missing_ = StatVar(**param_dict['stats']['missing'])
        obj.unique_ = StatVar(**param_dict['stats']['unique'])
        # distribution
        obj.vcounts_ = pd.Series(
            param_dict['distribution']['vcounts'], 
            name = obj.colname_
        )
        obj.vpercs_ = pd.Series(
            param_dict['distribution']['vpercs'], 
            name = obj.colname_
        )


    def generate(self, size: int = 100, seed: int = None, ignore_na:bool = False) -> pd.Series:
        if self.fitted_:
            rng = np.random.default_rng(seed = seed) # random generator
            choices = self.vpercs_.index.tolist()
            probs = self.vpercs_.values

            if not ignore_na:
                choices.append(pd.NA)
                probs = list(probs * (1 - self.missing_.perc))
                probs.append(self.missing_.perc)

            s = pd.Series(
                rng.choice(
                    a = choices,
                    size = size,
                    p = probs
                ),
                name = self.colname_,
            )
            if self.int_dtype_:
                s = s.astype('float').astype('Int64')
            return s
            
        else:
            raise Exception("Not fitted error")


class BinaryStat(CategStat):
    def __init__(self, pos_values: list, na_to_pos: bool = False, int_dtype: bool = False):
        super().__init__(int_dtype = int_dtype)
        self.pos_values_ = pos_values
        self.na_to_pos_ = na_to_pos

    def fit(self, vector: pd.Series):
        self.colname_ = vector.name
        # total number of records
        self.total_ = len(vector)
        # missing rate
        missing = vector.isna()
        self.missing_ = StatVar(
            value=serialize(missing.sum()), 
            perc=serialize(missing.mean())
        )
        # num of unique value (dropped NA value)
        unique = len(vector.dropna().unique())
        assert unique <= 2, f"Labeled as binary value, but {unique} categories are found"
        self.unique_ = StatVar(
            value=serialize(unique), 
            perc=serialize(unique / self.total_)
        )
        # category counts without null
        self.vcounts_ = vector.value_counts(
            dropna = True,
            normalize = False
        ).fillna(0)
        self.vpercs_ = (self.vcounts_ / self.vcounts_.sum()).fillna(0)

        # binary stat
        vector_binary = vector.copy()
        vector_binary[~vector_binary.isna()] = vector_binary[~vector_binary.isna()].apply(
            lambda v: 1 if v in self.pos_values_ else 0
        )
        vector_binary = vector_binary.fillna(int(self.na_to_pos_))
        self.binary_vcounts_ = vector_binary.value_counts(
            dropna = True,
            normalize = False
        ).fillna(0)
        self.binary_vpercs_ = (self.binary_vcounts_ / self.binary_vcounts_.sum()).fillna(0)

        self.fitted_ = True

    def collect_attrs(self) -> dict:
        # gather serializable parameters for __init__
        return dict(
            pos_values = self.pos_values_,
            int_dtype = self.int_dtype_,
            na_to_pos = self.na_to_pos_
        )

    def collect_params(self) -> dict:
        # gather serializable estimated parameters
        return dict(
            stats = dict(
                total = self.total_,
                missing = self.missing_._asdict(),
                unique = self.unique_._asdict()
            ),
            distribution = dict(
                vcounts = self.vcounts_.to_dict(),
                vpercs = self.vpercs_.to_dict(),
            ),
            binary_distribution = dict(
                binary_vcounts = self.binary_vcounts_.to_dict(),
                binary_vpercs = self.binary_vpercs_.to_dict(),
            )
        )

    @classmethod
    def set_params(cls, obj, param_dict: dict):
        # set params estimated to the new cls object
        # stats
        obj.total_ = param_dict['stats']['total']
        obj.missing_ = StatVar(**param_dict['stats']['missing'])
        obj.unique_ = StatVar(**param_dict['stats']['unique'])
        # distribution
        obj.vcounts_ = pd.Series(
            param_dict['distribution']['vcounts'], 
            name = obj.colname_
        )
        obj.vpercs_ = pd.Series(
            param_dict['distribution']['vpercs'], 
            name = obj.colname_
        )
        obj.binary_vcounts_ = pd.Series(
            param_dict['binary_distribution']['binary_vcounts'], 
            name = obj.colname_
        )
        obj.binary_vpercs_ = pd.Series(
            param_dict['binary_distribution']['binary_vpercs'], 
            name = obj.colname_
        )

class OrdinalCategStat(CategStat):
    def __init__(self, categories: List[Union[str, int]], int_dtype: bool = False):
        super().__init__(int_dtype = int_dtype)
        self.categories_ = categories

    def collect_attrs(self) -> dict:
        # gather serializable parameters for __init__
        return dict(
            int_dtype = self.int_dtype_,
            categories = self.categories_
        )
    
class NominalCategStat(CategStat):
    def to_ordinal(self, categories: List[Union[str, int]] = None) -> OrdinalCategStat:
        oc = OrdinalCategStat(
            categories = self.vcounts_.index.tolist() if categories is None else categories,
            int_dtype = self.int_dtype_
        )
        oc.colname_ = self.colname_
        self.set_params(
            self,
            param_dict = self.collect_params()
        )
        oc.fitted_ = self.fitted_
        return oc

def getNumericalStat(vector) -> Union[QuantileStat, DescStat]:
    vector = vector.astype('float')
    stat_quantile = QuantileStat(
        minimum = serialize(vector.min()),
        perc_1th = serialize(vector.quantile(0.01)),
        perc_5th = serialize(vector.quantile(0.05)),
        q1 = serialize(vector.quantile(0.25)),
        median = serialize(vector.quantile(0.5)),
        q3 = serialize(vector.quantile(0.75)),
        perc_95th = serialize(vector.quantile(0.95)),
        perc_99th = serialize(vector.quantile(0.99)),
        maximum = serialize(vector.max()),
        iqr = serialize(vector.quantile(0.75) - vector.quantile(0.25)),
        range = serialize(vector.max() - vector.min())
    )
    # descriptive statistics
    stat_descriptive = DescStat(
        mean = serialize(vector.mean()),
        var = serialize(vector.var()),
        std = serialize(vector.std()),
        skew = serialize(vector.skew()),
        kurt = serialize(vector.kurtosis()),
        mad = serialize(vector.mad()),
        cv = serialize(variation(vector, nan_policy='omit')),
        normality_p = serialize(kstest(vector.dropna(), cdf = 'norm')[1])
    )
    return stat_quantile, stat_descriptive

def getXtremeStat(vector, xtreme_method: Literal['iqr', 'quantile'] = 'iqr') -> XtremeStat:
    vector = vector.astype('float')
    if xtreme_method == 'iqr':
        q1 = vector.quantile(0.25)
        q3 = vector.quantile(0.75)
        iqr = q3 - q1
        lbound = q1 - 1.5 * iqr
        rbound = q3 + 1.5 * iqr
    
    elif xtreme_method == 'quantile':
        lbound = vector.quantile(0.01)
        rbound = vector.quantile(0.99)

    lvector = vector[vector < lbound]
    rvector = vector[vector > rbound]

    return XtremeStat(
        lbound = serialize(lbound),
        rbound = serialize(rbound),
        lxtreme_mean = serialize(lvector.mean()),
        lxtreme_median = serialize(lvector.median()),
        rxtreme_mean = serialize(rvector.mean()),
        rxtreme_median = serialize(rvector.median()),
    )

def log10pc(x):
    return np.where(x > 0, log10(x + 1), -log10(1 - x))

def exp10pc(x):
    return np.where(x > 0, exp10(x) - 1, -exp10(-x) + 1)

class NumericStat(_BaseStat):
    def __init__(self, setaside_zero: bool = False, log_scale: bool = False, xtreme_method: Literal['iqr', 'quantile'] = None, bins: int = 100):
        super().__init__()
        self.setaside_zero_ = setaside_zero
        self.log_scale_ = log_scale
        self.bins_ = bins
        self.xtreme_method_ = xtreme_method
        self.fitted_ = False

    def fit(self, vector: pd.Series):
        self.colname_ = vector.name
        # total number of records
        self.total_ = len(vector)
        # missing rate
        missing = vector.isna()
        self.missing_ = StatVar(
            value=serialize(missing.sum()), 
            perc=serialize(missing.mean())
        )
        # zero rate
        zeros = vector[vector == 0].count()
        self.zeros_ = StatVar(
            value=serialize(zeros), 
            perc=serialize(zeros / self.total_)
        )
        # infinite value
        infs_pos = vector[vector == np.inf].count()
        self.infs_pos_ = StatVar(
            value=serialize(infs_pos), 
            perc=serialize(infs_pos / self.total_)
        )
        infs_neg = vector[vector == -np.inf].count()
        self.infs_neg_ = StatVar(
            value=serialize(infs_neg), 
            perc=serialize(infs_neg / self.total_)
        )

        vector = vector.dropna()
        vector = vector[(vector < np.inf) & (vector > -np.inf)]
        if self.setaside_zero_:
            vector = vector[vector != 0]
        
        # xtreme value
        if self.xtreme_method_:
            self.xtreme_stat_ = getXtremeStat(vector, xtreme_method=self.xtreme_method_)
            vector_clean = vector[(vector >= self.xtreme_stat_.lbound) & (vector <= self.xtreme_stat_.rbound)]
            
        else:
            vector_clean = vector.copy()

        num_xtreme = len(vector) - len(vector_clean)
        self.xtreme_ = StatVar(
            value=serialize(num_xtreme), 
            perc=serialize(num_xtreme / self.total_)
        )

        # percentile & descriptive statistics
        self.stat_quantile_, self.stat_descriptive_= getNumericalStat(vector_clean)
        # histogram
        self.hist_, self.bin_edges_ = np.histogram(vector_clean, bins = self.bins_)

        if self.log_scale_:
            vector_log = pd.Series(log10pc(vector), name = self.colname_, dtype = 'float')

            # xtreme value
            if self.xtreme_method_:
                self.xtreme_stat_log_ = getXtremeStat(vector_log, xtreme_method=self.xtreme_method_)
                vector_clean_log = vector_log[(vector_log >= self.xtreme_stat_log_.lbound) & (vector_log <= self.xtreme_stat_log_.rbound)]
            else:
                vector_clean_log = vector_log.copy()

            num_xtreme_log = len(vector_log) - len(vector_clean_log)
            self.xtreme_log_ = StatVar(
                value=serialize(num_xtreme_log), 
                perc=serialize(num_xtreme_log / self.total_)
            )
            # percentile & descriptive statistics
            self.stat_quantile_log_, self.stat_descriptive_log_= getNumericalStat(vector_clean_log)
            # histogram
            self.hist_log_, self.bin_edges_log_ = np.histogram(vector_clean_log, bins = self.bins_)

        self.fitted_ = True

    def collect_attrs(self) -> dict:
        # gather serializable parameters for __init__
        return dict(
            setaside_zero = self.setaside_zero_,
            log_scale = self.log_scale_,
            bins = self.bins_,
            xtreme_method = self.xtreme_method_
        )

    def collect_params(self) -> dict:
        # gather serializable estimated parameters
        return dict(
            stats = dict(
                total = self.total_,
                missing = self.missing_._asdict(),
                zeros = self.zeros_._asdict(),
                infs_pos = self.infs_pos_._asdict(),
                infs_neg = self.infs_neg_._asdict(),
            ),
            xtreme_num = dict(
                origin = self.xtreme_._asdict() if hasattr(self, 'xtreme_') else None,
                log = self.xtreme_log_._asdict() if hasattr(self, 'xtreme_log_') else None
            ), # number of extreme values
            xtreme_stat = dict(
                origin = self.xtreme_stat_._asdict() if hasattr(self, 'xtreme_stat_') else None,
                log = self.xtreme_stat_log_._asdict() if hasattr(self, 'xtreme_stat_log_') else None
            ), # extreme value statistics
            stat_quantile = dict(
                origin = self.stat_quantile_._asdict(),
                log = self.stat_quantile_log_._asdict() if hasattr(self, 'stat_quantile_log_') else None
            ),
            stat_descriptive = dict(
                origin = self.stat_descriptive_._asdict(),
                log = self.stat_descriptive_log_._asdict() if hasattr(self, 'stat_descriptive_log_') else None
            ),
            histogram = dict(
                origin = dict(
                    hist = self.hist_.tolist(),
                    bin_edges = self.bin_edges_.tolist()
                ),
                log = dict(
                    hist = self.hist_log_.tolist() if hasattr(self, 'hist_log_') else None,
                    bin_edges = self.bin_edges_log_.tolist() if hasattr(self, 'bin_edges_log_') else None
                )
            )
        )

    @classmethod
    def set_params(cls, obj, param_dict: dict):
        # set params estimated to the new cls object
        obj.total_ = param_dict['stats']['total']
        obj.missing_ = StatVar(**param_dict['stats']['missing'])
        obj.zeros_ = StatVar(**param_dict['stats']['zeros'])
        obj.infs_pos_ = StatVar(**param_dict['stats']['infs_pos'])
        obj.infs_neg_ = StatVar(**param_dict['stats']['infs_neg'])
        obj.xtreme_ = StatVar(**param_dict['xtreme_num']['origin'])
        # origin statistics
        if obj.xtreme_method_ is not None:
            obj.xtreme_stat_ = XtremeStat(**param_dict['xtreme_stat']['origin'])
        obj.stat_quantile_ = QuantileStat(**param_dict['stat_quantile']['origin'])
        obj.stat_descriptive_ = DescStat(**param_dict['stat_descriptive']['origin'])
        obj.hist_ = np.array(param_dict['histogram']['origin']['hist'])
        obj.bin_edges_ = np.array(param_dict['histogram']['origin']['bin_edges'])
        # log statistics
        if obj.log_scale_:
            obj.xtreme_log_ = StatVar(**param_dict['xtreme_num']['log'])
            if obj.xtreme_method_ is not None:
                obj.xtreme_stat_log_ = XtremeStat(**param_dict['xtreme_stat']['log'])
            obj.stat_quantile_log_ = QuantileStat(**param_dict['stat_quantile']['log'])
            obj.stat_descriptive_log_ = DescStat(**param_dict['stat_descriptive']['log'])
            obj.hist_log_ = np.array(param_dict['histogram']['log']['hist'])
            obj.bin_edges_log_ = np.array(param_dict['histogram']['log']['bin_edges'])

    def generate(self, size: int = 100, seed: int = None, ignore_na:bool = False) -> pd.Series:
        if self.fitted_:
            if ignore_na:
                na_num = 0 # number of na values
                valid_num = size
                na_vs = []
            else:
                na_num = int(size * self.missing_.perc) # number of na values
                valid_num = size - na_num
                na_vs = [pd.NA] * na_num

            # size of zeros
            if self.setaside_zero_:
                zero_num = int(valid_num * self.zeros_.perc) # number of zeros
            else:
                zero_num = 0
            zero_vs = np.zeros(shape = zero_num)
            # size of infs
            inf_num = int(valid_num * self.infs_pos_.perc) # number of +inf
            neg_inf_num = int(valid_num * self.infs_neg_.perc) # number of -inf
            inf_vs = [np.inf] * inf_num
            neg_inf_vs = [-np.inf] * neg_inf_num

            # remaining valid nums
            valid_num = valid_num - zero_num - inf_num - neg_inf_num

            # valid numerical values
            if self.log_scale_:
                hist_dist = rv_histogram(
                    histogram=(self.hist_log_, self.bin_edges_log_),
                    seed = seed
                )
                valid_vs = exp10pc(hist_dist.rvs(size = valid_num))
            else:
                hist_dist = rv_histogram(
                    histogram=(self.hist_, self.bin_edges_),
                    seed = seed
                )
                valid_vs = hist_dist.rvs(size = valid_num)

            values = np.concatenate((valid_vs, inf_vs, neg_inf_vs, zero_vs, na_vs))
            np.random.shuffle(values)

            return pd.Series(
                values,
                name = self.colname_,
            ).astype(dtype='Float64')
        else:
            raise Exception("Not fitted error")


    
######### Univaraite Feature-Target correlation

def _combine_x_y(x: pd.Series, y: pd.Series, dropna: bool = True, combine_x_categ:bool = False) -> pd.DataFrame:
    df = pd.DataFrame({'x' : x.values, 'y' : y.values})
    if dropna:
        df = df.dropna(axis = 0, how = 'any')
    # categorical variable for x only:
    if combine_x_categ:
        vc = df['x'].value_counts(normalize = True)
        others = vc.iloc[20:].index  # only keep top 20 categories
        df.loc[df['x'].isin(others), 'x'] = 'Others'
    
    return df

class _BaseUniVarClfTargetCorr:
    def get_meta(self, x: pd.Series, y: pd.Series):
        self.colname_ = x.name
        self.yname_ = y.name
        self.ylabels_ = y.unique().astype('str').tolist()
    
    def fit(self, x: pd.Series, y: pd.Series):
        raise NotImplementedError("")
    
    def to_dict(self) -> dict:
        raise NotImplementedError("")
    
    @classmethod
    def from_dict(cls, d: dict):
        raise NotImplementedError("")

class CategUniVarClfTargetCorr(_BaseUniVarClfTargetCorr):
    def __init__(self, ):
        self.fitted_ = False
    
    def fit(self, x: pd.Series, y: pd.Series):
        self.get_meta(x, y)
        
        df = _combine_x_y(
            x.astype('str'), 
            y.astype('str'), 
            dropna = True, 
            combine_x_categ = True # combine x's categories to be less than 20
        )
        
        # categorical:  p(x | y)
        self.p_x_y_ = {}
        for y in df['y'].unique():
            cs = df.loc[df['y'] == y, ['x']].groupby('x').size()
            d = pd.DataFrame({
                    'count' : cs.values, 
                    'perc' : cs  / cs.sum()
                }, 
                index = cs.index
            )
            self.p_x_y_[y] = d.fillna(0) #.to_dict(orient = 'index')
            
        # categorical:  p(y | x)  prob (event rate when binary clf) by category
        self.p_y_x_ = {}
        for x in df['x'].unique():
            cs = df.loc[df['x'] == x, ['y']].groupby('y').size()
            d = pd.DataFrame({
                    'count' : cs.values, 
                    'perc' : cs  / cs.sum()
                }, 
                index = cs.index
            )
            self.p_y_x_[x] = d.fillna(0) #.to_dict(orient = 'index')
            
        self.fitted_ = True
            
    def to_dict(self) -> dict:
        if self.fitted_:
            return dict(
                colname = self.colname_,
                yname = self.yname_,
                ylabels = self.ylabels_,
                attr = dict(),
                p_x_y = {
                    k: v.to_dict(orient = 'index')
                    for k, v in self.p_x_y_.items() 
                },
                p_y_x = {
                    k: v.to_dict(orient = 'index')
                    for k, v in self.p_y_x_.items() 
                },
            )
        else:
            raise Exception("Not fitted error")
        
    @classmethod
    def from_dict(cls, d: dict):
        c = cls(
            **d['attr']
        )
        c.colname_ = d['colname']
        c.yname_ = d['yname']
        c.ylabels_ = d['ylabels']
        
        c.p_x_y_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in d['p_x_y'].items()
        }
        c.p_y_x_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in d['p_y_x'].items()
        }
        return c

class NumericUniVarClfTargetCorr(_BaseUniVarClfTargetCorr):
    def __init__(self, ):
        self.fitted_ = False
    
    def fit(self, x: pd.Series, y: pd.Series):
        self.get_meta(x, y)
        
        df = _combine_x_y(
            x.astype('float'), 
            y.astype('str'), 
            dropna = False,
        )
        
        # numerical:  p(x | y)  distribution by target (boxplot)
        t = {'origin' : {}, 'log' : {}}
        for y in df['y'].unique():
            v = df.loc[df['y'] == y, 'x']
            ns = NumericStat(
                setaside_zero=False, 
                log_scale=True, 
                xtreme_method='iqr', 
                bins=100
            )
            ns.fit(v)
            
            t['origin'][y] = {
                'lbound': ns.xtreme_stat_.lbound,
                'q1' : ns.stat_quantile_.q1,
                'mean' : ns.stat_descriptive_.mean,
                'median' : ns.stat_quantile_.median,
                'q3' : ns.stat_quantile_.q3,
                'rbound': ns.xtreme_stat_.rbound,
            }
            t['log'][y] = {
                'lbound': ns.xtreme_stat_log_.lbound,
                'q1' : ns.stat_quantile_log_.q1,
                'mean' : ns.stat_descriptive_log_.mean,
                'median' : ns.stat_quantile_log_.median,
                'q3' : ns.stat_quantile_log_.q3,
                'rbound': ns.xtreme_stat_log_.rbound,
            }
            
        self.p_x_y_ = {
            'origin' : pd.DataFrame.from_dict(t['origin'], orient='index'),
            'log' : pd.DataFrame.from_dict(t['log'], orient='index')
        }
            
            
        # numerical:  p(y | x)  prob (event rate if binary clf) by bucketized x
        mob = MulticlassOptimalBinning(
            prebinning_method = 'quantile',
            monotonic_trend = 'auto',
            max_n_prebins = 20,
            min_prebin_size = 0.05,
        )
        mob.fit(x = df['x'], y = df['y'])
        self.p_y_x_ = mob.binning_table.build(add_totals=False)
        
        self.fitted_ = True
        
    def to_dict(self) -> dict:
        if self.fitted_:
            return dict(
                colname = self.colname_,
                yname = self.yname_,
                ylables = self.ylabels_,
                attr = dict(),
                p_x_y = {
                    k: v.to_dict(orient = 'index')
                    for k, v in self.p_x_y_.items() 
                },
                p_y_x = (
                    self.p_y_x_
                    .replace({np.nan: None})
                    .to_dict(orient = 'records')
                ),
            )
        else:
            raise Exception("Not fitted error")
        
    @classmethod
    def from_dict(cls, d: dict):
        c = cls(
            **d['attr']
        )
        c.colname_ = d['colname']
        c.yname_ = d['yname']
        c.ylabels_ = d['ylabels']
        
        c.p_x_y_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in d['p_x_y'].items()
        }
        c.p_y_x_ = pd.DataFrame.from_records(d['p_y_x'])
        return c


#########  Tabular

class _BaseTabularSerializable:
    def to_dict(self) -> dict:
        rs = {}
        for col, config in self.configs.items():
            r = config.to_dict()
            r['constructor'] = config.__class__.__name__
            rs[col] = r
        return rs
    
    @classmethod
    def from_dict(cls, r: dict):
        configs = {}
        for col, c in r.items():
            constructor = globals().get(c['constructor'])
            config = constructor.from_dict(c)
            configs[col] = config
        
        return cls(configs = configs)

    def dump(self, js_path: str):
        js = self.to_dict()
        toJSON(js = js, file = js_path)

    @classmethod
    def load(cls, js_path):
        js = loadJSON(file = js_path)
        return cls.from_dict(js)


class TabularStat(_BaseTabularSerializable):
    def __init__(self, configs: Dict[str, _BaseStat], n_jobs:int = 1):
        self.configs = configs
        self.n_jobs = n_jobs
        
    def get_categ_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, CategStat) or schema.classname in ('BinaryStat','NominalCategStat','OrdinalCategStat')]
    
    def get_nominal_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, NominalCategStat) or schema.classname == 'NominalCategStat']
    
    def get_binary_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, BinaryStat) or schema.classname == 'BinaryStat']
    
    def get_ordinal_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, OrdinalCategStat) or schema.classname == 'OrdinalCategStat']

    def get_numeric_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, NumericStat) or schema.classname == 'NumericStat']

    def fit(self, df: pd.DataFrame):
        # support multi-processor running
        cols = df.columns.intersection(self.configs.keys(), sort=False)
        jobs = (self._fit_one(col, df[col]) for col in cols)
        stats = parallel_run(jobs, n_jobs = self.n_jobs)
        self.configs = {col: stat for col, stat in zip(cols, stats)}

    @delayer
    def _fit_one(self, col: str, vector: pd.Series) -> _BaseStat:
        stat = self.configs[col]
        stat.fit(vector)
        return stat
    
    def generate(self, size: int = 100, seed: int = None, ignore_na:bool = False) -> pd.DataFrame:
        return pd.concat(
            [config.generate(
                size = size,
                seed = seed,
                ignore_na = ignore_na
            ) for col, config in self.configs.items()],
            axis = 1
        )


class TabularUniVarClfTargetCorr(_BaseTabularSerializable):
    def __init__(self, configs: Dict[str, _BaseUniVarClfTargetCorr], n_jobs:int = 1):
        self.configs = configs
        self.n_jobs = n_jobs
        
    def get_categ_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, CategUniVarClfTargetCorr)]

    def get_numeric_cols(self) -> List[str]:
        return [col for col, schema in self.configs.items() if isinstance(schema, NumericUniVarClfTargetCorr)]

    def fit(self, X: pd.DataFrame, y: pd.Series):
        # support multi-processor running
        cols = X.columns.intersection(self.configs.keys(), sort=False)
        jobs = (self._fit_one(col, X[col], y) for col in cols)
        stats = parallel_run(jobs, n_jobs = self.n_jobs)
        self.configs = {col: stat for col, stat in zip(cols, stats)}

    @delayer
    def _fit_one(self, col: str, x: pd.Series, y: pd.Series) -> _BaseUniVarClfTargetCorr:
        stat = self.configs[col]
        stat.fit(x, y)
        return stat